import re
from typing import Dict, List, Union


class KonsepyMatch:
    """Wrapper for re.Match that handles duplicate named groups."""

    def __init__(self, match: re.Match, group_mapping: Dict[str, List[str]]):
        self._match = match
        self._group_mapping = group_mapping

    def group(self, *args):
        """Proxy group() and handle synthesized names."""
        if not args:
            return self._match.group(0)

        results = []
        for arg in args:
            if isinstance(arg, str) and arg in self._group_mapping:
                # find first internal group that matched
                found = False
                for internal_name in self._group_mapping[arg]:
                    val = self._match.group(internal_name)
                    if val is not None:
                        results.append(val)
                        found = True
                        break
                if not found:
                    results.append(None)
            else:
                results.append(self._match.group(arg))

        if len(results) == 1:
            return results[0]
        return tuple(results)

    def groups(self, default=None):
        """Proxy groups()."""
        return self._match.groups(default)

    def groupdict(self, default=None):
        """Proxy groupdict() and collapse duplicate names."""
        raw_dict = self._match.groupdict(default)
        collapsed = {}
        for original_name, internal_names in self._group_mapping.items():
            for internal_name in internal_names:
                val = raw_dict.get(internal_name)
                if val is not None:
                    collapsed[original_name] = val
                    break
            else:
                collapsed[original_name] = default
        return collapsed

    def start(self, group=0):
        """Proxy start()."""
        if isinstance(group, str) and group in self._group_mapping:
            for internal_name in self._group_mapping[group]:
                if self._match.group(internal_name) is not None:
                    return self._match.start(internal_name)
        return self._match.start(group)

    def end(self, group=0):
        """Proxy end()."""
        if isinstance(group, str) and group in self._group_mapping:
            for internal_name in self._group_mapping[group]:
                if self._match.group(internal_name) is not None:
                    return self._match.end(internal_name)
        return self._match.end(group)

    def span(self, group=0):
        """Proxy span()."""
        if isinstance(group, str) and group in self._group_mapping:
            for internal_name in self._group_mapping[group]:
                if self._match.group(internal_name) is not None:
                    return self._match.span(internal_name)
        return self._match.span(group)

    def __getattr__(self, name):
        """Forward any other attributes to the original match object."""
        return getattr(self._match, name)

    def __str__(self):
        data = {
            'match': self.group(0),
            'span': self.span(),
        }
        groups = self.groupdict()
        if groups:
            data['groups'] = groups
        args = ', '.join(f'{k}={v!r}' for k, v in data.items())
        return f'KonsepyMatch({args})'

    def __repr__(self):
        return self.__str__()


class KonsepyRegex:
    """Wrapper for compiled regex that handles optional duplicate named groups."""

    def __init__(self, pattern: Union[str, re.Pattern], flags: int = 0, allow_dupe_names: bool = True):
        self._original_pattern = pattern.pattern if isinstance(pattern, re.Pattern) else pattern
        self._flags = flags
        self._group_mapping = {}
        if allow_dupe_names and isinstance(pattern, str):
            # find all (?P<name>...)
            named_group_re = re.compile(r'\(\?P<(?P<name>[a-zA-Z_][a-zA-Z0-9_]*)>')

            group_counts = {}

            def rename_match(m):
                name = m.group('name')
                if name not in group_counts:
                    group_counts[name] = 1
                    internal_name = name
                else:
                    group_counts[name] += 1
                    internal_name = f'{name}__dup{group_counts[name]}'

                if name not in self._group_mapping:
                    self._group_mapping[name] = []
                self._group_mapping[name].append(internal_name)

                return f'(?P<{internal_name}>'

            pattern = named_group_re.sub(rename_match, pattern)

        if isinstance(pattern, str):
            self._pattern = re.compile(pattern, flags)
        else:
            self._pattern = pattern

    def finditer(self, string: str, pos: int = 0, endpos: int = 2147483647):
        """Proxy finditer() and wrap results."""
        for m in self._pattern.finditer(string, pos, endpos):
            yield KonsepyMatch(m, self._group_mapping) if self._group_mapping else m

    def search(self, string: str, pos: int = 0, endpos: int = 2147483647):
        """Proxy search() and wrap result."""
        m = self._pattern.search(string, pos, endpos)
        if not m:
            return None
        return KonsepyMatch(m, self._group_mapping) if self._group_mapping else m

    def match(self, string: str, pos: int = 0, endpos: int = 2147483647):
        """Proxy match() and wrap result."""
        m = self._pattern.match(string, pos, endpos)
        if not m:
            return None
        return KonsepyMatch(m, self._group_mapping) if self._group_mapping else m

    def fullmatch(self, string: str, pos: int = 0, endpos: int = 2147483647):
        """Proxy fullmatch() and wrap result."""
        m = self._pattern.fullmatch(string, pos, endpos)
        if not m:
            return None
        return KonsepyMatch(m, self._group_mapping) if self._group_mapping else m

    def __getattr__(self, name):
        """Forward any other attributes to the original pattern object."""
        return getattr(self._pattern, name)

    def __str__(self):
        group_names = list(self._group_mapping.keys()) if self._group_mapping else []
        return (
            f'KonsepyRegex(pattern={self._original_pattern!r}, '
            f'groups={group_names!r})'
        )

    def __repr__(self):
        return (
            f'KonsepyRegex('
            f'pattern={self._original_pattern!r}, '
            f'compiled={self._pattern.pattern!r}, '
            f'flags={self._flags!r}, '
            f'group_mapping={self._group_mapping!r})'
        )


def rx_compile(pattern: str, flags: int = 0) -> KonsepyRegex:
    r"""
    Compile a regex pattern, allowing duplicate named groups in alternation branches.

    Example:
        compile_pattern_allow_dupe_names(r'(?:score: (?P<val>\d+)|results: (?P<val>\d+))')
    """
    return KonsepyRegex(pattern, flags=flags, allow_dupe_names=True)


class RxType(type):
    MAPPING = {
        'p': '.',
        'w': r'\w',
        'W': r'\W',
        'S': r'\S',
        's': r'\s',
        'o': r'(?:\w+\W*)',  # word
    }

    def __getattr__(cls, item):
        return cls._parse_item(item)

    def _parse_item(cls, element):
        el = cls.MAPPING[element[0]]
        if len(element) > 1:
            return el + cls._parse_numbers(*element[1:].split('_'))
        return el

    def _parse_numbers(cls, *nums):
        if len(nums) == 1:
            v1 = 0
            v2 = nums[0]
        else:
            v1, v2 = nums
        return rf'{{{v1},{v2}}}'


class Rx(metaclass=RxType):
    pass



import datetime
import json
from collections import Counter
from pathlib import Path

from loguru import logger

from konsepy.bio_tag_sentence import get_pipeline
from konsepy.cli import add_outdir_and_infiles
from konsepy.constants import NOTEDATE_LABEL, ID_LABEL, NOTEID_LABEL, NOTETEXT_LABEL
from konsepy.textio import iterate_csv_file


def get_target_overlap(sentences: list[tuple[int, str]], max_previous):
    """Determine how many sentences can be retained to provide additional context."""
    if len(sentences) == 1:
        return 2, [], 0  # if only a single sentence, don't include in next window
    total_length = 0
    total_char_length = 0
    for i, (length, sentence) in enumerate(sentences[::-1]):
        if total_length + length >= max_previous:
            if i == 0 and length > max_previous * 2:
                return 2, [], 0  # previous sentence is too long
            return length + 2, sentences[min(-1, -i):], total_char_length
        total_length += length
        total_char_length += len(sentence)
    return total_length + 2, sentences[-1:], total_char_length  # just the last sentence


def corpus2jsonl(input_files, outdir: Path, *,
                 split=None, max_seq_length=512, tokenizer=None,
                 sentence_model='senter',
                 id_label=ID_LABEL, noteid_label=NOTEID_LABEL,
                 notedate_label=NOTEDATE_LABEL, notetext_label=NOTETEXT_LABEL,
                 noteorder_label=None, metadata_labels=None, encoding='utf8',
                 limit_noteids=None):
    """
    Convert a corpus to a jsonl format, suitable for using prodigy
    """
    if split in {'sentence', 'sent_chunk', 'sent_window'}:
        nlp = get_pipeline(sentence_model)
    if split in {'sent_chunk', 'sent_window'}:
        from transformers import AutoTokenizer
        tokenizer = AutoTokenizer.from_pretrained(tokenizer)
    start_time = datetime.datetime.now()
    dt = start_time.strftime("%Y%m%d_%H%M%S")
    if split is not None:
        name = f'corpus_{dt}.{split}.jsonl'
    else:
        name = f'corpus_{dt}.jsonl'
    length_counter = Counter()
    with open(outdir / name, 'w', encoding='utf8') as out:
        for count, studyid, note_id, note_date, text, metadata in iterate_csv_file(
                input_files, encoding=encoding,
                id_label=id_label, noteid_label=noteid_label,
                notedate_label=notedate_label, notetext_label=notetext_label,
                noteorder_label=noteorder_label, metadata_labels=metadata_labels,
        ):
            if limit_noteids and note_id not in limit_noteids:
                continue
            if split is None:
                out.write(json.dumps({
                    'id': count,
                    'studyid': studyid,
                    'note_id': note_id,
                    'note_date': note_date,
                    'start_index': 0,
                    'end_index': len(text),
                    'text': text,
                }) + '\n')
            elif split == 'sentence':
                doc = nlp(text)
                for sent_id, spacy_sentence in enumerate(doc.sents):
                    sentence = str(spacy_sentence)
                    out.write(json.dumps({
                        'id': count,
                        'studyid': studyid,
                        'note_id': note_id,
                        'note_date': note_date,
                        'sentence_id': sent_id,
                        'start_index': spacy_sentence.start_char,
                        'end_index': spacy_sentence.end_char,
                        'text': sentence,
                    }) + '\n')
            elif split == 'section':
                raise NotImplementedError
            elif split == 'sent_chunk':
                doc = nlp(text)
                sents = list(doc.sents)
                chunk_id = 0
                curr = []
                curr_length = 2  # start/end
                start_index = 0  # approximate character index from start
                end_index = None
                for sent_id, spacy_sentence in enumerate(sents):
                    sentence = str(spacy_sentence)
                    length = len(tokenizer.tokenize(sentence))  # if len(sent) > max_seq_length, allow truncation
                    if length + curr_length > max_seq_length:
                        # TODO: what if curr_length == 0 (i.e., len(sentence) > max_seq_length)?
                        out.write(json.dumps({
                            'id': count,
                            'studyid': studyid,
                            'note_id': note_id,
                            'note_date': note_date,
                            'chunk_id': chunk_id,
                            'start_index': start_index,
                            'end_index': end_index,
                            'text': ' '.join(curr),  # sentences retain their punctuation chars, but not spaces
                        }) + '\n')
                        # reset vars
                        chunk_id += 1
                        curr = [sentence]
                        curr_length = 2 + length
                        start_index = spacy_sentence.start_char
                        end_index = spacy_sentence.end_char
                    else:
                        length_counter[length] += 1
                        curr.append(sentence)
                        curr_length += length
                        end_index = spacy_sentence.end_char
                        if spacy_sentence[-1].is_space and spacy_sentence[-1].text.count('\n') > 1:
                            # end of a section/paragraph
                            out.write(json.dumps({
                                'id': count,
                                'studyid': studyid,
                                'note_id': note_id,
                                'note_date': note_date,
                                'chunk_id': chunk_id,
                                'start_index': start_index,
                                'end_index': end_index,
                                'text': ' '.join(curr),  # sentences retain their punctuation chars
                            }) + '\n')
                            # reset vars
                            chunk_id += 1
                            curr = []
                            curr_length = 2
                            start_index = spacy_sentence.end_char
                # final fencepost
                if curr:
                    out.write(json.dumps({
                        'id': count,
                        'studyid': studyid,
                        'note_id': note_id,
                        'note_date': note_date,
                        'chunk_id': chunk_id,
                        'start_index': start_index,
                        'end_index': end_index,
                        'text': ' '.join(curr),  # sentences retain their punctuation chars
                    }) + '\n')
            elif split == 'sent_window':
                target_overlap = int(max_seq_length / 4)
                doc = nlp(text)
                sents = list(doc.sents)
                chunk_id = 0
                curr = []  # tuple(length: int, sentence: str)
                curr_length = 2  # start/end
                start_index = 0
                end_index = None
                for sent_id, spacy_sentence in enumerate(sents):
                    sentence = str(spacy_sentence)
                    length = len(tokenizer.tokenize(sentence))  # if len(sent) > max_seq_length, allow truncation
                    if length + curr_length > max_seq_length:
                        out.write(json.dumps({
                            'id': count,
                            'studyid': studyid,
                            'note_id': note_id,
                            'note_date': note_date,
                            'chunk_id': chunk_id,
                            'start_index': start_index,
                            'end_index': end_index,
                            'text': ' '.join(c for _, c in curr),  # sentences retain their punctuation chars
                        }) + '\n')
                        # reset vars
                        chunk_id += 1
                        curr_length, curr, char_length = get_target_overlap(curr, target_overlap)
                        curr.append((length, sentence))
                        curr_length = 2 + length
                        start_index = spacy_sentence.start_char - char_length  # this is the next sentence - overlap
                        end_index = spacy_sentence.end_char
                    else:
                        length_counter[length] += 1
                        curr.append((length, sentence))
                        curr_length += length
                        end_index = spacy_sentence.end_char
                        if spacy_sentence[-1].is_space and spacy_sentence[-1].text.count('\n') > 1:
                            # end of a section/paragraph
                            out.write(json.dumps({
                                'id': count,
                                'studyid': studyid,
                                'note_id': note_id,
                                'note_date': note_date,
                                'chunk_id': chunk_id,
                                'start_index': start_index,
                                'end_index': end_index,
                                'text': ' '.join(c for _, c in curr),  # sentences retain their punctuation chars
                            }) + '\n')
                            # reset vars
                            chunk_id += 1
                            curr_length, curr, char_length = get_target_overlap(curr, target_overlap)
                            start_index = spacy_sentence.end_char + 1 - char_length  # this sentence - overlap
                # final fencepost
                if curr:
                    out.write(json.dumps({
                        'id': count,
                        'studyid': studyid,
                        'note_id': note_id,
                        'note_date': note_date,
                        'chunk_id': chunk_id,
                        'start_index': start_index,
                        'end_index': end_index,
                        'text': ' '.join(c for _, c in curr),  # sentences retain their punctuation chars
                    }) + '\n')
            else:
                raise ValueError(f'Unrecognized option for splitting: {split}')
            if count % 50000 == 0:
                logger.info(f'Completed {count:,} records ({datetime.datetime.now()})')
    end_time = datetime.datetime.now()
    if length_counter:
        logger.info(f'Too long so truncated: {len(length_counter)}')
        logger.info(f' * {sorted(length_counter, reverse=True)[:5]}')
    logger.info(f'DONE: Completed {count:,} records ({end_time})')
    logger.info(f'DONE: Total processing time: {end_time - start_time}.')


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(fromfile_prefix_chars='@!')
    add_outdir_and_infiles(parser)
    parser.add_argument('--split', default=None,
                        choices=['sentence', 'section', 'sent_chunk', 'sent_window'],
                        help='How to split/divide the text. Defaults to None (i.e., keep entire text together).')
    parser.add_argument('--max-seq-length', dest='max_seq_length', default=512, type=int,
                        help='Number of tokens to retain when `--split` is sent_chunk or sent_window.')
    parser.add_argument('--tokenizer', default=None,
                        help='Path to BERT tokenizer.')
    corpus2jsonl(**vars(parser.parse_args()))



import argparse
import datetime
import json
from pathlib import Path
try:
    from datasets import Dataset, Sequence
except ImportError:
    Dataset = None
    Sequence = None
from konsepy.cli import clean_args
from loguru import logger


def get_spans(spans):
    """Get start, end, and domain (i.e., tag)"""
    for span in spans:
        yield span['start'], span['end'], span['domain'], False


def update_word(i, tokens, ner_tags, curr_word, span_start, span_is_middle, span_tag):
    tokens.append(''.join(curr_word))
    if i >= span_start:
        if span_is_middle:
            ner_tags.append(f'I-{span_tag}')
        else:
            ner_tags.append(f'B-{span_tag}')
            span_is_middle = True
    else:
        ner_tags.append('O')
    return span_is_middle


def create_bio_dataset(path: Path, outpath: Path, test_size=0.1, validation_size=0.05, note_id_field='note_id'):
    """Convert jsonl file into a datasets.DatasetDict for use in model building.

    path: path to jsonl file generated by `bio_tag.py`
    outpath: path to place dataset
    test_size: size of test set (default: 0.05)
    validation_size: size of validation set (default: 0.1)
    note_id_field: name of field containing unique identifier to the note/document (default: note_id)
    """
    all_note_ids = []
    all_tokens = []
    all_ner_tags = []
    with open(path, encoding='utf8') as fh:
        for line in fh:
            data = json.loads(line.strip())
            tokens = []
            ner_tags = []
            spans = get_spans(data['results'])
            span_start, span_end, span_tag, span_is_middle = next(spans)

            curr_word = []
            for i, letter in enumerate(data['text']):
                if letter == ' ':  # end of word
                    if curr_word:
                        span_is_middle = update_word(
                            i, tokens, ner_tags, curr_word, span_start,
                            span_is_middle, span_tag,
                        )
                        curr_word = []
                elif letter.isalpha():
                    curr_word.append(letter)
                else:  # punctuation
                    if curr_word:
                        span_is_middle = update_word(
                            i, tokens, ner_tags, curr_word, span_start,
                            span_is_middle, span_tag,
                        )
                        curr_word = []
                    tokens.append(letter)
                    ner_tags.append('O')
                if i >= span_end:
                    try:
                        span_start, span_end, span_tag, span_is_middle = next(spans)
                    except StopIteration:
                        span_start, span_end = 100_000, 100_000
            if curr_word:
                span_is_middle = update_word(
                    i, tokens, ner_tags, curr_word, span_start,
                    span_is_middle, span_tag,
                )
            all_note_ids.append(data[note_id_field])
            all_tokens.append(tokens)
            all_ner_tags.append(ner_tags)

    tagset = set([x for el in all_ner_tags for x in el])
    logger.info(f'Tagset of length {len(tagset)}: {",".join(tagset)}')

    ds = Dataset.from_dict({
        'tokens': Sequence(features=all_tokens),
        'ner_tags': Sequence(features=all_ner_tags),
        'note_ids': all_note_ids,
    })

    res = ds.train_test_split(validation_size, shuffle=True)
    length = len(all_note_ids)
    num = length - (length * validation_size)
    denom = length * test_size
    ds = res['train'].train_test_split(test_size=num / denom, shuffle=True)
    ds['validation'] = res['test']
    ds.save_to_disk(outpath / f'{path.stem}.{datetime.datetime.now().strftime("%Y%m%d_%H%M%S")}.dataset')


def create_bio_dataset_args():
    parser = argparse.ArgumentParser(fromfile_prefix_chars='@')
    parser.add_argument('path', type=Path,
                        help='Path to jsonl input file.')
    parser.add_argument('outpath', type=Path,
                        help='Path to write created dataset.')
    parser.add_argument('--test-size', dest='test_size', type=float, default=0.1,
                        help='Percent of data to use for testing.')
    parser.add_argument('--validation-size', dest='validation_size', type=float, default=0.05,
                        help='Percent of data to hold out for final validation.')
    parser.add_argument('--note-id-field', dest='note_id_field', type=str, default='note_id',
                        help='Specify label in jsonl file for distinguishing unique documents/notes.')
    create_bio_dataset(**clean_args(vars(parser.parse_args())))


if __name__ == '__main__':
    create_bio_dataset_args()



import json
from pathlib import Path
from konsepy.run_all_matches import run_all_matches


def test_run_all_matches_direct(tmp_path):
    input_file = Path('test/data/corpus.jsonl')
    outdir = tmp_path / 'direct_run'
    outdir.mkdir()

    result_dir = run_all_matches(
        input_files=[str(input_file)],
        outdir=outdir,
        package_name='example_nlp',
        id_label='chapter',
        noteid_label='chapter'
    )

    assert result_dir.exists()
    assert result_dir.name.startswith('run_all_matches_')

    output_jsonl = result_dir / 'output.jsonl'
    assert output_jsonl.exists()

    with open(output_jsonl, encoding='utf8') as f:
        lines = f.readlines()
        assert len(lines) > 0

        for line in lines:
            data = json.loads(line)
            assert 'studyid' in data
            assert 'note_id' in data
            assert 'concept' in data
            assert 'category' in data
            assert 'precontext' in data
            assert 'match' in data
            assert 'postcontext' in data
            assert 'start_index' in data
            assert 'end_index' in data
            # check that it's one match per row (in run_all it was 'categories' list)
            assert 'categories' not in data
            assert isinstance(data['category'], str)


def test_run_all_matches_date_dupe_named_groups(tmp_path):
    input_file = tmp_path / "date_input.jsonl"
    input_file.write_text(
        (
            '{"chapter": "1", "text": "It was the 24th of February, 1815. '
            'Another letter was dated February 25, 1815."}\n'
        ),
        encoding="utf8",
    )

    outdir = tmp_path / "out"
    outdir.mkdir()

    result_dir = run_all_matches(
        input_files=[str(input_file)],
        outdir=outdir,
        package_name="misc_nlp",
        concepts=["date"],
        id_label="chapter",
        noteid_label="chapter",
        notetext_label="text",
    )

    rows = []
    with open(result_dir / "output.jsonl", encoding="utf8") as f:
        for line in f:
            rows.append(json.loads(line))

    date_rows = [row for row in rows if row["concept"] == "date"]
    assert len(date_rows) >= 2

    assert any(
        row["match"] == "24th of February, 1815"
        and row.get("day") == "24"
        and row.get("month").lower() == "february"
        and row.get("year") == "1815"
        for row in date_rows
    )

    assert any(
        row["match"] == "February 25, 1815"
        and row.get("day") == "25"
        and row.get("month").lower() == "february"
        and row.get("year") == "1815"
        for row in date_rows
    )



import re


def get_precontext(m, text, window=20, start=None, word_window=None):
    if start is None:
        start = m.start()
    if word_window:
        # for simplicity and correctness, we search backwards from start
        prefix = text[:start]
        # find word boundaries
        words = re.findall(r'\w+', prefix)
        if len(words) > word_window:
            # find the start index of the nth word from the end
            matches = list(re.finditer(r'\w+', prefix))
            if len(matches) >= word_window:
                target_match = matches[-word_window]
                return prefix[target_match.start():]
        return prefix

    return text[max(0, start - window): start]


def get_postcontext(m, text, window=20, end=None, word_window=None):
    if end is None:
        end = m.end()
    if word_window:
        suffix = text[end:]
        matches = list(re.finditer(r'\w+', suffix))
        if len(matches) >= word_window:
            target_match = matches[word_window - 1]
            return suffix[:target_match.end()]
        return suffix

    return text[end: end + window]


def get_around(m, text, window=20, start=None, end=None, word_window=None):
    if start is None:
        start = m.start()
    if end is None:
        end = m.end()
    if word_window:
        pre = get_precontext(m, text, start=start, word_window=word_window)
        post = get_postcontext(m, text, end=end, word_window=word_window)
        return pre + text[start:end] + post

    return text[max(0, start - window): end + window]


def get_contexts(m, text, window=20, context_match=None, context_window=None, context_direction=0, word_window=None):
    """
    if matching on context of a previous match, the offsets will need to be updated. To do this, use:
        * context_match: previous match
        * context_window: previous window (supplied by `window` parameter)
        * context_direction: -1 for previous context, 1 for post-context
    """
    if context_match:
        if context_direction == -1:
            offset = max(0, context_match.start() - context_window)
            start = offset + m.start()
            end = offset + m.end()
        elif context_direction == 1:
            offset = context_match.end()
            start = offset + m.start()
            end = offset + m.end()
        else:
            raise ValueError(fr'Unrecognized context direction: {context_direction};'
                             ' expected -1 [before] or +1 [after]')
        precontext = get_precontext(m, text, window, start=start, word_window=word_window)
        postcontext = get_postcontext(m, text, window, end=end, word_window=word_window)
        around = get_around(m, text, window, start=start, end=end, word_window=word_window)
    else:
        precontext = get_precontext(m, text, window, word_window=word_window)
        postcontext = get_postcontext(m, text, window, word_window=word_window)
        around = get_around(m, text, window, word_window=word_window)
    return {
        'm': m,
        'precontext': precontext,
        'postcontext': postcontext,
        'text': text,
        'window': window,
        'word_window': word_window,
        'around': around,
    }


def check_if_pattern_after(pattern, m, text, window=20, banned_characters='.', end=None, return_concept=None, **kwargs):
    postcontext = get_postcontext(m, text, window=window, end=end)
    if m2 := pattern.search(postcontext):
        for ch in banned_characters:
            if ch in postcontext[:m2.start()]:
                return None
        return m2 if return_concept is None else return_concept


def check_if_pattern_before(pattern, m, text, window=20, banned_characters='.', start=None, return_concept=None, **kwargs):
    precontext = get_precontext(m, text, window=window, start=start)
    if m2 := pattern.search(precontext):
        for ch in banned_characters:
            if ch in precontext[m2.end():]:
                return None
        return m2 if return_concept is None else return_concept


def check_if_pattern_around(pattern, m, text, window=20, banned_characters='.', start=None, end=None, return_concept=None, **kwargs):
    around = get_around(m, text, window, start=start, end=end)
    for m2 in pattern.finditer(around):
        if m2.end() < m.start():  # match before
            for ch in banned_characters:
                if ch in around[m2.end():m.start()]:
                    continue
        elif m2.start() > m.end():
            for ch in banned_characters:
                if ch in around[m.end():m2.start()]:
                    continue
        return m2 if return_concept is None else return_concept



"""
Get biotags for a corpus.

Usage: python bio_tag.py --outdir /path/to/output --infiles infile.csv infile.sas7bdat

Post-processing:
* Identify which `count`/index values should be removed based on noise rules
* Remove these from the output jsonl file

"""

import csv
import json
from pathlib import Path

from konsepy.cli import add_outdir_and_infiles, concept_cli
from konsepy.constants import NOTETEXT_LABEL, NOTEDATE_LABEL, NOTEID_LABEL, ID_LABEL
from konsepy.importer import get_all_concepts
from konsepy.types import RegexDict


def build_regex_dict(package_name, concepts):
    if not package_name:
        raise ValueError('Specify name of package.')
    return {concept.name: concept.run_func for concept in get_all_concepts(package_name, *concepts or list())}


def get_bio_tags_corpus(input_files, outdir: Path, *, package_name: str = None,
                        concepts=None, regexes: RegexDict = None,
                        id_label=ID_LABEL, noteid_label=NOTEID_LABEL,
                        notedate_label=NOTEDATE_LABEL, notetext_label=NOTETEXT_LABEL,
                        noteorder_label=None, **kwargs):
    """

    """
    outdir.mkdir(exist_ok=True)

    if not package_name and not regexes:
        raise ValueError(f'Either `regexes` or `package_name` must be specified.')
    regexes = regexes or build_regex_dict(package_name, concepts)

    fieldnames = ['index', 'id', 'studyid', 'note_id', 'chunk_id', 'domain', 'category', 'capture', 'start', 'end']

    with (
        open(outdir / 'bio_tag_data.corpus.csv', 'w', newline='') as out,
        open(outdir / 'bio_tag_data.corpus.jsonl', 'w') as jsonl,
    ):
        writer = csv.DictWriter(out, fieldnames=fieldnames)
        writer.writeheader()
        i = 0
        for jsonl_file in input_files:
            with open(jsonl_file, encoding='utf8') as fh:
                for line in fh:
                    data = json.loads(line.strip())
                    extras = {'chunk_start': data['start_index']}
                    if 'sentence_id' in data:
                        extras['chunk_id'] = data['sentence_id']
                    data['results'] = []
                    for domain, regex_func in regexes.items():
                        for category, (capture, start, end) in zip(regex_func(data['text'])):
                            curr_data = {
                                'index': i,
                                'domain': domain,
                                'category': category,
                                'capture': capture,
                                'start': start,
                                'end': end,
                            }
                            writer.writerow(
                                {k: v for k, v in (data | curr_data | extras).items()
                                 if k in fieldnames})
                            data['results'].append(curr_data)
                            i += 1
                        # write 1 line per input sentence
                        data['results'] = sorted(data['results'], key=lambda x: x['start'])
                        jsonl.write(json.dumps(data) + '\n')


if __name__ == '__main__':
    concept_cli(get_bio_tags_corpus)



import re

import pytest

from konsepy.context.negation import has_prenegation, has_postnegation, has_negation, is_not_negated

_preneg = [
    # (pattern, text, banned, exp_pre, exp_any)
    ('cancer', 'no cancer', '.', 'no', 'no'),
    ('cancer', 'cancer: no', '.', None, 'no'),
    ('cancer', 'smoking: no. Cancer diagnosis', '.', None, None),  # intervening banned char
    ('cancer', 'smoking: no. Cancer diagnosis', '', 'no', 'no'),  # remove banned char
]

_postneg = [
    # (pattern, text, banned, exp_post, exp_any)
    ('cancer', 'no cancer', '.', None, 'no'),
    ('cancer', 'cancer absent', '.', 'absent', 'absent'),
    ('cancer', 'cancer: no', ':', None, None),  # intervening banned char
    ('cancer', 'cancer: no', '.', 'no', 'no'),  # remove banned char
]

preneg = [x[0:4] for x in _preneg]  # remove exp_any
postneg = [x[0:4] for x in _postneg]  # remove exp_any
# looks for both, but from either pre-context or post context of match
allneg = [x[0:3] + (x[4],) for x in _preneg + _postneg]
# looks from left end or right end of the text (regardless of the match)
anyneg = [x[:3] for x in _preneg + _postneg]


@pytest.mark.parametrize('pattern, text, banned_characters, exp', preneg)
def test_has_prenegation(pattern, text, banned_characters, exp):
    m = re.search(pattern, text, re.I)
    if res := has_prenegation(text[:m.start()], banned_characters=banned_characters):
        assert res.group() == exp
    else:
        assert res is None


@pytest.mark.parametrize('pattern, text, banned_characters, exp', preneg)
def test_has_negation_neg1_direction(pattern, text, banned_characters, exp):
    m = re.search(pattern, text, re.I)
    if res := has_negation(text[:m.start()], direction=-1, banned_characters=banned_characters):
        assert res.group() == exp
    else:
        assert res is None


@pytest.mark.parametrize('pattern, text, banned_characters, exp', postneg)
def test_has_postnegation(pattern, text, banned_characters, exp):
    m = re.search(pattern, text, re.I)
    if res := has_postnegation(text[m.end():], banned_characters=banned_characters):
        assert res.group() == exp
    else:
        assert res is None


@pytest.mark.parametrize('pattern, text, banned_characters, exp', postneg)
def test_has_postnegation(pattern, text, banned_characters, exp):
    m = re.search(pattern, text, re.I)
    if res := has_negation(text[m.end():], direction=1, banned_characters=banned_characters):
        assert res.group() == exp
    else:
        assert res is None


@pytest.mark.parametrize('pattern, text, banned_characters, exp', allneg)
def test_has_negation(pattern, text, banned_characters, exp):
    m = re.search(pattern, text, re.I)
    if res := has_negation(text, direction=0, banned_characters=banned_characters, m=m):
        assert res.group() == exp
    else:
        assert res is None


@pytest.mark.parametrize('pattern, text, banned_characters, exp', allneg)
def test_not_negated(pattern, text, banned_characters, exp):
    m = re.search(pattern, text, re.I)
    res = is_not_negated(m, text[:m.start()], text[m.end():], banned_characters=banned_characters)
    assert res is not bool(exp)


@pytest.mark.parametrize('pattern, text, banned_characters', anyneg)
def test_any_negation(pattern, text, banned_characters):
    m = re.search(pattern, text, re.I)
    res = has_negation(text, direction=0, banned_characters=banned_characters)
    assert res is not None



import re

import pytest

from konsepy.context.other_subject import has_other_subject, check_if_other_subject


@pytest.mark.parametrize('pattern, text, banned_characters, exp', [
    ('drinks?', 'father drinks', '.', 'father'),
    ('drinks?', 'drinks mother', '.', None),
])
def test_other_subject_before(pattern, text, banned_characters, exp):
    m = re.search(pattern, text, re.I)
    precontext = text[:m.start()]
    if res := has_other_subject(precontext, -1, banned_characters=banned_characters):
        assert res.group() == exp
    else:
        assert res == exp


@pytest.mark.parametrize('pattern, text, banned_characters, exp', [
    ('drinks?', 'father drinks', '.', None),
    ('drinks?', 'drinks mother', '.', 'mother'),
])
def test_other_subject_after(pattern, text, banned_characters, exp):
    m = re.search(pattern, text, re.I)
    postcontext = text[m.end():]
    if res := has_other_subject(postcontext, 1, banned_characters=banned_characters):
        assert res.group() == exp
    else:
        assert res == exp


@pytest.mark.parametrize('pattern, text, banned_characters, exp', [
    ('drinks?', 'father drinks', '.', 'father'),
    ('drinks?', 'drinks mother', '.', 'mother'),
    ('drinks?', 'drinks at mother\'s home', '.', None),
    ('drinks?', 'according to father, drinks', '.', None),
])
def test_check_if_other_subject(pattern, text, banned_characters, exp):
    m = re.search(pattern, text, re.I)
    precontext = text[:m.start()]
    postcontext = text[m.end():]
    if res := check_if_other_subject(m, precontext, postcontext, text):
        assert res.group() == exp
    else:
        assert res == exp  # None



"""
Usage: python bio_tag.py --outdir /path/to/output --infiles infile.csv infile.sas7bdat

Post-processing:
* Identify which `count`/index values should be removed based on noise rules
* Remove these from the output jsonl file

"""

import csv
import json
from pathlib import Path

from loguru import logger

from konsepy.cli import add_outdir_and_infiles, concept_cli
from konsepy.constants import NOTETEXT_LABEL, NOTEDATE_LABEL, NOTEID_LABEL, ID_LABEL
from konsepy.engine import ProcessingEngine
from konsepy.types import RegexDict


def get_bio_tags(input_files, outdir: Path, *, package_name: str = None,
                 id_label=ID_LABEL, noteid_label=NOTEID_LABEL,
                 notedate_label=NOTEDATE_LABEL, notetext_label=NOTETEXT_LABEL,
                 noteorder_label=None, metadata_labels=None,
                 concepts=None, **kwargs):
    """
    Generate BIO tagged data using ProcessingEngine.
    """
    outdir.mkdir(exist_ok=True)

    if not package_name:
        raise ValueError('`package_name` must be specified.')

    engine = ProcessingEngine(
        input_files, package_name, id_label=id_label, noteid_label=noteid_label,
        notedate_label=notedate_label, notetext_label=notetext_label,
        noteorder_label=noteorder_label, metadata_labels=metadata_labels,
        concepts=concepts, **kwargs
    )

    with (
        open(outdir / 'bio_tag_data.csv', 'w', newline='') as out,
        open(outdir / 'bio_tag_data.jsonl', 'w') as jsonl,
    ):
        writer = csv.DictWriter(
            out,
            fieldnames=['index', 'studyid', 'note_id', 'domain', 'category', 'capture', 'start', 'end']
        )
        writer.writeheader()
        
        state = {'index': 0}

        def callback(studyid, note_id, note_date, text, metadata, concept, categories, matches):
            constant_meta = {
                'studyid': studyid,
                'note_id': note_id,
            }
            curr_note = {
                'results': [],
            }
            if matches:
                for category, m in zip(categories, matches):
                    data = {
                        'index': state['index'],
                        'domain': concept.name,
                        'category': category,
                        'capture': m.group(),
                        'start': m.start(),
                        'end': m.end(),
                    }
                    writer.writerow(data | constant_meta)
                    curr_note['results'].append(data)
                    state['index'] += 1
            
            # NOTE: The original code wrote to JSONL inside the concept loop, which seems wrong
            # as it would write one line per concept per note.
            # ProcessingEngine's callback is called per concept per note.
            # However, the original code had:
            # for domain, regex_func in regexes.items():
            #     ...
            #     jsonl.write(json.dumps(curr_note) + '\n')
            # This means it was indeed writing one line per concept per note.
            
            curr_note['results'] = sorted(curr_note['results'], key=lambda x: x['start'])
            jsonl.write(json.dumps(curr_note) + '\n')

        engine.run(callback)


if __name__ == '__main__':
    concept_cli(get_bio_tags)



import json
from konsepy.run_all_matches import run_all_matches

def test_run_all_matches_with_dupe_names(tmp_path):
    # create a mock corpus file with content that matches both branches
    corpus_file = tmp_path / 'corpus.jsonl'
    with open(corpus_file, 'w', encoding='utf8') as f:
        f.write(json.dumps({'chapter': '1', 'notetext': 'Score: 123. Results: 456.'}) + '\n')
    
    outdir = tmp_path / 'out'
    outdir.mkdir()

    result_dir = run_all_matches(
        input_files=[str(corpus_file)],
        outdir=outdir,
        package_name='misc_nlp',
        concepts=['dupe_names'],
        id_label='chapter',
        noteid_label='chapter',
        notetext_label='notetext'
    )

    output_jsonl = result_dir / 'output.jsonl'
    assert output_jsonl.exists()

    with open(output_jsonl, encoding='utf8') as f:
        lines = f.readlines()
        assert len(lines) == 2
        
        # first match (score: 123)
        data1 = json.loads(lines[0])
        assert data1['match'].lower() == 'score: 123'
        assert data1['val'] == '123'
        assert 'val__dup2' not in data1
        
        # second match (results: 456)
        data2 = json.loads(lines[1])
        assert data2['match'].lower() == 'results: 456'
        assert data2['val'] == '456'
        assert 'val__dup2' not in data2



import json
import re
from pathlib import Path

import pytest

from konsepy.run4snippets import run4snippets


@pytest.fixture
def input_file(request):
    return Path(request.fspath).parent / 'data' / 'corpus.jsonl'


def iter_jsonl_output(outdir: Path, extra_keys=None):
    extra_keys = set(extra_keys) if extra_keys else set()
    exp_keys = {'note_id', 'concept', 'category', 'studyid', 'note_date', 'match', 'start_index',
                'end_index', 'precontext', 'postcontext', 'pretext', 'posttext'} | extra_keys
    with open(outdir / 'output.jsonl', encoding='utf8') as fh:
        for line in fh:
            data = json.loads(line)
            assert set(data.keys()) == exp_keys
            yield data


def test_run4snippets(tmp_path, input_file, caplog):
    outdir = run4snippets(
        [input_file],
        outdir=tmp_path,
        package_name='example_nlp',
        id_label='chapter',
        noteid_label='chapter',
    )
    # test log results
    loglines = [line for line in caplog.text.split('\n') if line.strip()]
    assert any('Arguments ignored: {}' in line for line in loglines)
    assert any(re.search(r'Loaded \d+ concepts for processing', line) for line in loglines)
    assert any('Output 713 rows' in line for line in loglines)
    assert any('Total records: 117' in line for line in loglines)

    # test jsonlines output
    for data in iter_jsonl_output(outdir):
        pass


def test_run4snippets_with_metadata(tmp_path, input_file, caplog):
    metadata = ['chapter', 'title']
    outdir = run4snippets(
        [input_file],
        outdir=tmp_path,
        package_name='example_nlp',
        id_label='chapter',
        noteid_label='chapter',
        order_metadata=metadata,
        metadata_labels=metadata,
    )
    # test log results
    loglines = [line for line in caplog.text.split('\n') if line.strip()]
    assert any('Arguments ignored: {}' in line for line in loglines)
    assert any('Loaded 3 concepts for processing' in line for line in loglines)
    assert any('Output 713 rows' in line for line in loglines)
    assert any('Total records: 117' in line for line in loglines)

    # test jsonlines output
    for data in iter_jsonl_output(outdir, extra_keys=set(metadata)):
        assert metadata[0] in data.keys()  # confirm presence
        assert metadata[0] == list(data.keys())[0]  # ensure this is in the correct order
        assert metadata[1] in data.keys()  # confirm presence
        assert metadata[1] == list(data.keys())[1]  # ensure this is in the correct order


def test_run4snippets_only_categories(tmp_path, input_file, caplog):
    target_categories = ['Jealous.NO', 'Revenge.NO', 'Justice.NO']
    outdir = run4snippets(
        [input_file],
        outdir=tmp_path,
        package_name='example_nlp',
        id_label='chapter',
        noteid_label='chapter',
        target_categories=target_categories,
    )
    # test log results
    loglines = [line for line in caplog.text.split('\n') if line.strip()]
    assert any('Arguments ignored: {}' in line for line in loglines)
    assert any('Loaded 3 concepts for processing' in line for line in loglines)
    assert any('Output 4 rows' in line for line in loglines)
    assert any('Total records: 117' in line for line in loglines)

    # test jsonlines output
    for data in iter_jsonl_output(outdir):
        assert data['category'] in target_categories


def test_run4snippets_only_concepts(tmp_path, input_file, caplog):
    target_concepts = ['justice']
    outdir = run4snippets(
        [input_file],
        outdir=tmp_path,
        package_name='example_nlp',
        id_label='chapter',
        noteid_label='chapter',
        target_concepts=target_concepts,
    )
    # test log results
    loglines = [line for line in caplog.text.split('\n') if line.strip()]
    assert any('Arguments ignored: {}' in line for line in loglines)
    assert any('Loaded 3 concepts for processing' in line for line in loglines)
    assert any('Output 536 rows' in line for line in loglines)
    assert any('Total records: 117' in line for line in loglines)

    # test jsonlines output
    for data in iter_jsonl_output(outdir):
        assert data['concept'] in target_concepts


def test_run4snippets_only_concepts_and_categories(tmp_path, input_file, caplog):
    target_concepts = ['revenge']
    target_categories = ['Jealous.FAMILY']

    outdir = run4snippets(
        [input_file],
        outdir=tmp_path,
        package_name='example_nlp',
        id_label='chapter',
        noteid_label='chapter',
        target_concepts=target_concepts,
        target_categories=target_categories,
    )
    # test log results
    loglines = [line for line in caplog.text.split('\n') if line.strip()]
    assert any('Arguments ignored: {}' in line for line in loglines)
    assert any('Loaded 3 concepts for processing' in line for line in loglines)
    assert any('Output 135 rows' in line for line in loglines)
    assert any('Total records: 117' in line for line in loglines)

    # test jsonlines output
    for data in iter_jsonl_output(outdir):
        assert data['concept'] in target_concepts or data['category'] in target_categories



import re
from konsepy.rxutils import rx_compile, KonsepyRegex


def test_regex_wrapper_direct_no_dupes():
    pattern = r'(?P<val>\d+)'
    # test without allow_dupe_names 
    rw = KonsepyRegex(pattern, allow_dupe_names=False)
    m = rw.search('123')
    assert m is not None
    assert m.group('val') == '123'
    # should be raw re.Match if no group_mapping
    assert isinstance(m, re.Match)


def test_regex_wrapper_direct_with_dupes():
    pattern = r'(?:(?P<val>A)|(?P<val>B))'
    # test with allow_dupe_names (standard behavior)
    rw = KonsepyRegex(pattern)
    m1 = rw.search('A')
    assert m1.group('val') == 'A'
    m2 = rw.search('B')
    assert m2.group('val') == 'B'
    # should be KonsepyMatch
    from konsepy.rxutils import KonsepyMatch
    assert isinstance(m1, KonsepyMatch)


def test_regex_wrapper_with_compiled_pattern():
    pattern = re.compile(r'(?P<val>\d+)')
    rw = KonsepyRegex(pattern)
    m = rw.search('123')
    assert m.group('val') == '123'
    assert isinstance(m, re.Match)


def test_duplicate_named_groups_alternation():
    pattern = r'(?:score: (?P<val>\d+)|results: (?P<val>\d+))'
    compiled = rx_compile(pattern, re.I)

    # test first branch
    m1 = compiled.search('score: 123')
    assert m1 is not None
    assert m1.group(0) == 'score: 123'
    assert m1.group('val') == '123'
    assert m1.groupdict() == {'val': '123'}
    assert m1.start('val') == 7
    assert m1.end('val') == 10

    # test second branch
    m2 = compiled.search('results: 456')
    assert m2 is not None
    assert m2.group(0) == 'results: 456'
    assert m2.group('val') == '456'
    assert m2.groupdict() == {'val': '456'}
    assert m2.start('val') == 9
    assert m2.end('val') == 12


def test_finditer_returns_wrapped_matches():
    pattern = r'(?:(?P<val>A)|(?P<val>B))'
    compiled = rx_compile(pattern)
    text = 'A B'
    matches = list(compiled.finditer(text))
    assert len(matches) == 2
    assert matches[0].group('val') == 'A'
    assert matches[1].group('val') == 'B'
    assert matches[0].groupdict() == {'val': 'A'}
    assert matches[1].groupdict() == {'val': 'B'}


def test_normal_pattern_behavior():
    pattern = r'(?P<first>\w+) (?P<second>\w+)'
    compiled = rx_compile(pattern)
    m = compiled.search('hello world')
    assert m.group('first') == 'hello'
    assert m.group('second') == 'world'
    assert m.groupdict() == {'first': 'hello', 'second': 'world'}


def test_multiple_duplicates():
    pattern = r'(?P<v>1)|(?P<v>2)|(?P<v>3)'
    compiled = rx_compile(pattern)

    m1 = compiled.search('1')
    assert m1.group('v') == '1'

    m2 = compiled.search('2')
    assert m2.group('v') == '2'

    m3 = compiled.search('3')
    assert m3.group('v') == '3'


def test_group_positional_still_works():
    pattern = r'(?:(?P<val>A)|(?P<val>B))'
    compiled = rx_compile(pattern)
    m = compiled.search('B')
    # internal pattern is (?P<val>A)|(?P<val__dup2>B)
    # group 0: B
    # group 1: None (val)
    # group 2: B (val__dup2)
    assert m.group(0) == 'B'
    assert m.group(1) is None
    assert m.group(2) == 'B'
    assert m.groups() == (None, 'B')


def test_multiple_args_to_group():
    pattern = r'(?P<a>A)|(?P<a>B)'
    compiled = rx_compile(pattern)
    m = compiled.search('B')
    assert m.group(0, 'a') == ('B', 'B')


def test_konsepy_regex_str_shows_user_facing_pattern_and_groups():
    regex = KonsepyRegex(
        r'(?:score: (?P<val>\d+)|results: (?P<val>\d+))',
        flags=re.I,
    )

    text = str(regex)

    assert text.startswith('KonsepyRegex(')
    assert r"pattern='(?:score: (?P<val>\\d+)|results: (?P<val>\\d+))'" in text
    assert "groups=['val']" in text


def test_konsepy_regex_repr_shows_compiled_pattern_and_group_mapping():
    regex = KonsepyRegex(
        r'(?:score: (?P<val>\d+)|results: (?P<val>\d+))',
        flags=re.I,
    )

    text = repr(regex)

    assert text.startswith('KonsepyRegex(')
    assert r"pattern='(?:score: (?P<val>\\d+)|results: (?P<val>\\d+))'" in text
    assert r"compiled='(?:score: (?P<val>\\d+)|results: (?P<val__dup2>\\d+))'" in text
    assert 'flags=' in text
    assert "'val': ['val', 'val__dup2']" in text


def test_konsepy_regex_str_without_dupe_names_is_clear():
    regex = KonsepyRegex(
        r'(?P<word>\w+)',
        allow_dupe_names=True,
    )

    text = str(regex)

    assert text.startswith('KonsepyRegex(')
    assert r"pattern='(?P<word>\\w+)'" in text
    assert "groups=['word']" in text



import argparse
import json
from pathlib import Path

from loguru import logger

from transformers import AutoTokenizer, DataCollatorForTokenClassification, AutoModelForTokenClassification, \
    TrainingArguments, Trainer
from datasets import load_dataset, load_from_disk, DatasetDict, load_metric
import numpy as np


def tokenize_adjust_labels(tokenizer, label2id):
    def _inner_tokenize_adjust_labels(all_samples_per_split):
        """Get values for input_ids, token_type_ids, attention_mask"""
        tokenized_samples = tokenizer.batch_encode_plus(
            all_samples_per_split['tokens'],
            is_split_into_words=True,
            padding='max_length',
            truncation=True,
            max_length=512,
        )
        # tokenized_samples is not a datasets object so this alone won't work with Trainer API, hence map is used
        # so the new keys [input_ids, labels (after adjustment)]
        # can be added to the datasets dict for each train test validation split
        total_adjusted_labels = []
        for k in range(0, len(tokenized_samples['input_ids'])):
            prev_wid = -1
            word_ids_list = tokenized_samples.word_ids(batch_index=k)
            existing_label_ids = all_samples_per_split['ner_tags'][k]
            i = -1
            adjusted_label_ids = []

            for wid in word_ids_list:
                if wid is None:
                    adjusted_label_ids.append(-100)
                elif wid != prev_wid:
                    i = i + 1
                    adjusted_label_ids.append(label2id[existing_label_ids[i]])
                    prev_wid = wid
                else:
                    adjusted_label_ids.append(label2id[existing_label_ids[i]])
            total_adjusted_labels.append(adjusted_label_ids)
        tokenized_samples['labels'] = total_adjusted_labels
        return tokenized_samples

    return _inner_tokenize_adjust_labels


def compute_metrics(p, id2label):
    metric = load_metric('seqeval')
    predictions, labels = p
    predictions = np.argmax(predictions, axis=2)

    # Remove ignored index (special tokens)
    true_predictions = [
        [id2label[p] for (p, l) in zip(prediction, label) if l != -100]
        for prediction, label in zip(predictions, labels)
    ]
    true_labels = [
        [id2label[l] for (p, l) in zip(prediction, label) if l != -100]
        for prediction, label in zip(predictions, labels)
    ]

    results = metric.compute(predictions=true_predictions, references=true_labels)
    flattened_results = {
        'overall_precision': results['overall_precision'],
        'overall_recall': results['overall_recall'],
        'overall_f1': results['overall_f1'],
        'overall_accuracy': results['overall_accuracy'],
        'predicted_b': len([x for p in true_predictions for x in p if x.startswith('B')]),
        'predicted_i': len([x for p in true_predictions for x in p if x.startswith('I')]),
        'predicted_o': len([x for p in true_predictions for x in p if x.startswith('O')]),
    }


def train_on_bio_dataset(dataset_path: Path, outpath: Path, run_name: str, pretrained_tokenizer: str = None,
                         pretrained_model: str = 'emilyalsentzer/Bio_ClinicalBERT', **params):
    dataset = DatasetDict.load_from_disk(dataset_path)
    tagset = set([x for doc in dataset['train']['ner_tags'] for x in doc])
    label2id = {x: i for i, x in enumerate(tagset)}
    id2label = {v: k for k, v in label2id.items()}
    tokenizer = AutoTokenizer.from_pretrained(pretrained_tokenizer or pretrained_model)

    tokenized_dataset = dataset.map(tokenize_adjust_labels(tokenizer, label2id), batched=True)
    data_collator = DataCollatorForTokenClassification(
        tokenizer,
        padding='max_length',
        max_length=512,
    )
    model = AutoModelForTokenClassification.from_pretrained(pretrained_model)
    default_params = {
                         'evaluation_strategy': 'steps',
                         'learning_rate': 2e-4,
                         'per_device_train_batch_size': 32,
                         'per_device_eval_batch_size': 32,
                         'num_train_epochs': 50,
                         'weight_decay': 0.01,
                         'logging_steps': 1000,
                         'save_strategy': 'no',
                     } | params
    training_args = TrainingArguments(
        output_dir=str(outpath),
        run_name=run_name,
        **default_params,
    )
    trainer = Trainer(
        model=model,
        args=training_args,
        train_dataset=tokenized_dataset['train'],
        eval_dataset=tokenized_dataset['test'],
        data_collator=data_collator,
        tokenizer=tokenizer,
        compute_metrics=compute_metrics
    )
    trainer.train()
    trainer.save_model(outpath / f'{run_name}.model')
    with open(outpath / 'id2label.json', 'w') as out:
        json.dump(id2label, out)
    with open(outpath / 'label2id.json', 'w') as out:
        json.dump(label2id, out)

    val_results = trainer.predict(tokenized_dataset['validation'])
    logger.info(val_results)
    logger.info(f'Label-2-id: {label2id}')
    logger.info(f'Id-2-label: {id2label}')


def train_on_bio_dataset_args():
    parser = argparse.ArgumentParser(fromfile_prefix_chars='@')
    parser.add_argument('dataset_path', type=Path,
                        help='Path to dataset (output from `create_bio_dataset.py`)')
    parser.add_argument('outpath', type=Path,
                        help='Path to write model to.')
    parser.add_argument('run_name', type=str,
                        help='Name of this run.')
    parser.add_argument('--pretrained-model', dest='pretrained_model', type=str,
                        default='emilyalsentzer/Bio_ClinicalBERT',
                        help='Path to model to fine-tune, or to huggingface model.')
    parser.add_argument('--pretrained-tokenizer', dest='pretrained_tokenizer', type=str,
                        help='Path to tokenizer for model, or to huggingface tokenizer;'
                             ' defaults to same as `pretrained_model`')
    parser.add_argument('--param', nargs='*',
                        help='Parameters to supply to `TrainingArguments` in form of `num_train_epochs==50`')
    args = vars(parser.parse_args())
    params = {x.split('==')[0]: x.split('==')[1] for x in args['param']}
    del args['param']
    train_on_bio_dataset(**args, **params)


if __name__ == '__main__':
    train_on_bio_dataset_args()



from konsepy.context.contexts import get_contexts

_DEFAULT_WINDOW = 30


def search_first_regex(regexes, window=_DEFAULT_WINDOW, word_window=None):
    """
    For each regex in the list, return only the first match found in the text.

    Args:
        regexes: List of tuples (regex, category, *other).
        window: Context window size for functions.
        word_window: Size of context window in terms of words (overrides default window)

    Returns:
        A function that takes text and returns a generator of results.
    """

    def _search_first_regex(text, *, include_match=False):
        found = False
        for regex, category, *other in regexes:
            funcs = _unpack_other_funcs(other)
            for m in regex.finditer(text):
                found = True
                yield from _yield_categories(m, category, funcs, text, window,
                                             word_window=word_window, include_match=include_match)
                break
            if found:
                break

    return _search_first_regex


def search_all_regex(regexes, window=_DEFAULT_WINDOW, word_window=None):
    """
    For each regex in the list, return all matches found in the text.

    Args:
        regexes: List of tuples (regex, category, *other).
        window: Context window size for functions.
        word_window: Size of context window in terms of words (overrides default window)

    Returns:
        A function that takes text and returns a generator of results.
    """

    def _search_all_regex(text, *, include_match=False):
        for regex, category, *other in regexes:
            funcs = _unpack_other_funcs(other)
            for m in regex.finditer(text):
                yield from _yield_categories(m, category, funcs, text, window,
                                             word_window=word_window, include_match=include_match)

    return _search_all_regex


def get_all_regex_by_index(regexes, window=_DEFAULT_WINDOW, word_window=None):
    """
    For each regex, return all results along with their start and end indices.

    Args:
        regexes: List of tuples (regex, category, *other).
        window: Context window size for functions.
        word_window: Size of context window in terms of words (overrides default window)

    Returns:
        A function that takes text and returns a generator of (result, match_text, start, end).
    """

    def _get_all_regex_by_index(text):
        for regex, category, *other in regexes:
            funcs = _unpack_other_funcs(other)
            for m in regex.finditer(text):
                for res in _yield_categories(m, category, funcs, text, window, word_window=word_window):
                    if isinstance(res, (list, tuple)):
                        res, m2 = res
                        yield res, m2.group(), m2.start(), m2.end()
                    else:
                        yield res, m.group(), m.start(), m.end()

    return _get_all_regex_by_index


def search_all_regex_match_func(regexes, window=_DEFAULT_WINDOW, word_window=None):
    """
    For each regex, apply functions to the match object to determine the result.

    Args:
        regexes: List of tuples (regex, category, *other).
        window: Context window size for functions.
        word_window: Size of context window in terms of words (overrides default window)

    Returns:
        A function that takes text and returns a generator of results.
    """

    def _search_all_regex(text, *, include_match=False):
        for regex, category, *other in regexes:
            funcs = _unpack_other_funcs(other)
            for m in regex.finditer(text):
                yield from _yield_categories(m, category, funcs, text, window,
                                             word_window=word_window, include_match=include_match)

    return _search_all_regex


def _unpack_other_funcs(other):
    """
    Extract function(s) from optional arguments.

    Returns:
        List of functions or None.
    """
    if len(other) > 0:
        if isinstance(other[0], list):
            return other[0]
        else:
            return [other[0]]
    return None


def _yield_categories(m, category, funcs, text, window, *, word_window=None, include_match=False):
    """Helper to yield categories based on functions and match object."""
    if funcs:
        for func in funcs:
            if func and (res := func(**get_contexts(m, text, window, word_window=word_window))):
                if isinstance(res, list):
                    res, m = res  # category, match
                yield (res, m) if include_match else res
            else:
                yield (category, m) if include_match else category
    else:
        yield (category, m) if include_match else category


def search_and_replace_regex_func(regexes, window=_DEFAULT_WINDOW, word_window=None):
    """
    Search for regexes, but replace found text with dots to prevent double-matching.

    This is useful when one regex might be a subset of another or when you want to
    ensure each piece of text is only categorized once.

    Args:
        regexes: List of tuples (regex, category, *other).
        window: Context window size for functions.
        word_window: Size of context window in terms of words (overrides default window)

    Returns:
        A function that takes text and returns a generator of results.
    """

    def _search_all_regex(text, include_match=False):
        for regex, category, *other in regexes:
            funcs = _unpack_other_funcs(other)
            text_pieces = []
            prev_end = 0
            for m in regex.finditer(text):
                found = None
                if funcs:
                    for func in funcs:  # parse function in order
                        if res := func(**get_contexts(m, text, window, word_window=word_window)):
                            found = (res, m) if include_match else res
                            break
                if found:
                    if found is True or (include_match and found[0] is True):
                        continue  # no result
                    yield found
                else:
                    yield (category, m) if include_match else category
                text_pieces.append(text[prev_end:m.start()])
                text_pieces.append(f' {(len(m.group()) - 2) * "."} ')
                prev_end = m.end()
            text_pieces.append(text[prev_end:])
            text = ''.join(text_pieces)

    return _search_all_regex


def search_first_regex_func(regexes, window=_DEFAULT_WINDOW, word_window=None):
    """
    Return only the very first result found among all regexes.

    Args:
        regexes: List of tuples (regex, category, *other).
        window: Context window size for functions.
        word_window: Size of context window in terms of words (overrides default window)

    Returns:
        A function that takes text and returns a generator yielding at most one result.
    """

    def _search_first_regex(text, include_match=False):
        for regex, category, *other in regexes:
            funcs = _unpack_other_funcs(other)

            for m in regex.finditer(text):
                for res in _yield_categories(m, category, funcs, text, window,
                                             word_window=word_window, include_match=include_match):
                    yield res
                    return

    return _search_first_regex


def search_all_regex_func(regexes, window=_DEFAULT_WINDOW, word_window=None):
    """
    For each regex, return all matches, running any provided context functions.

    Setting a regex to None acts as a sentinel value: if any non-UNKNOWN regex matched
    before the sentinel, processing stops.

    Args:
        regexes: List of tuples (regex, category, *other).
        window: Context window size for functions.
        word_window: Size of context window in terms of words (overrides default window)

    Returns:
        A function that takes text and returns a generator of results.
    """

    def _search_all_regex(text, include_match=False, ignore_indices=False):
        """
        ignore_indices: set to True for testing (so that the locating index doesn't need to be run)
        """
        found_non_unknown = False
        for regex, category, *other in regexes:
            if regex is None:
                if found_non_unknown:
                    # setting regex to None acts as sentinel value: continue only if only UNKNOWNs were found
                    break
                else:
                    continue

            funcs = _unpack_other_funcs(other)
            if len(other) > 1 and not ignore_indices:  # function to locate starting text
                indices = list(other[1](text))
            else:
                indices = [(0, len(text))]

            for start, end in indices:
                for m in regex.finditer(text, pos=start, endpos=end):
                    if funcs:
                        found_any = False
                        found = False
                        for func in funcs:
                            if func is None:
                                continue
                            if res := func(**get_contexts(m, text, window, word_window=word_window)):
                                yield (res, m) if include_match else res
                                found = True
                                found_any = True
                                if hasattr(res, 'name') and res.name != 'UNKNOWN':
                                    found_non_unknown = True
                                elif not hasattr(res, 'name') and res != 'UNKNOWN':
                                    found_non_unknown = True
                            if found:
                                break
                        if not found_any and category is not None:
                            yield (category, m) if include_match else category
                            if hasattr(category, 'name') and category.name != 'UNKNOWN':
                                found_non_unknown = True
                            elif not hasattr(category, 'name') and category != 'UNKNOWN':
                                found_non_unknown = True
                    elif category is not None:
                        yield (category, m) if include_match else category
                        if hasattr(category, 'name') and category.name != 'UNKNOWN':
                            found_non_unknown = True
                        elif not hasattr(category, 'name') and category != 'UNKNOWN':
                            found_non_unknown = True

    return _search_all_regex



import re

from konsepy.context.contexts import get_contexts

DEFAULT_PRENEG_PAT = re.compile(
    rf'\b(?:family|or|no concerns?|no|none|deny|denies|denied|(?:no|neg|negative)\W*to|not)\b'
)
DEFAULT_POSTNEG_PAT = re.compile(
    rf'\b(?:no concerns?|or|no|none|deny|denies|denied|absent|not at all)\b'
)


def is_not_negated(m, precontext, postcontext, banned_characters='.', **kwargs):
    return not (
            has_prenegation(precontext, banned_characters=banned_characters)
            or has_postnegation(postcontext, banned_characters=banned_characters)
    )


def check_if_negated(m, precontext, postcontext, text, window, neg_concept=True, **kwargs):
    direction = 0
    if m2 := has_negation(precontext, direction=-1, **kwargs):
        direction = -1
    elif m2 := has_negation(postcontext, direction=1, **kwargs):
        direction = 1
    if m2:
        if is_not_negated(**get_contexts(
                m2, text, context_match=m, context_window=window, context_direction=direction
        )):
            return None
        return neg_concept
    return None


def has_negation(text, direction=0, prenegation_pat=DEFAULT_PRENEG_PAT,
                 postnegation_pat=DEFAULT_POSTNEG_PAT, banned_characters='.',
                 m=None, window=20, **kwargs):
    """Return first mention of negation, looking both backward or forward through the text.

    direction = set to -1 (precontext) or 1 (postcontext) to ensure is in same sentence
    """
    if m is not None:
        pretext = text[max(0, m.start() - window):m.start()]
        posttext = text[m.end(): m.end() + window]
    else:
        pretext = text
        posttext = text
    if direction == -1:  # precontext
        return has_prenegation(pretext, prenegation_pat, banned_characters)
    elif direction == 1:  # postcontext
        return has_postnegation(posttext, postnegation_pat, banned_characters)
    elif direction == 0:  # both directions
        return (has_prenegation(pretext, prenegation_pat, banned_characters)
                or has_postnegation(posttext, postnegation_pat, banned_characters))
    else:
        raise ValueError(f'Unexpected negation direction: {direction} (expected: -1 [pre], 0 [both], or 1 [post]).')


def has_prenegation(text, prenegation_pat=DEFAULT_PRENEG_PAT, banned_characters='.'):
    if m := prenegation_pat.search(text):
        for ch in banned_characters:
            if ch in text[m.end():]:
                return None
        return m


def has_postnegation(text, postnegation_pat=DEFAULT_POSTNEG_PAT, banned_characters='.'):
    if m := postnegation_pat.search(text):
        for ch in banned_characters:
            if ch in text[:m.start()]:
                return None
        return m



import argparse
import sys
from pathlib import Path
from loguru import logger

from konsepy.run_all import run_all
from konsepy.run_all_matches import run_all_matches
from konsepy.run4snippets import run4snippets
from konsepy.bio_tag import get_bio_tags
from konsepy.corpus2jsonl import corpus2jsonl
from konsepy.create_bio_dataset import create_bio_dataset
from konsepy.cli import add_outdir_and_infiles, add_run_all_args, clean_args, clean_metadata_labels


def main():
    parser = argparse.ArgumentParser(description='konsepy: NLP information extraction using regular expressions')
    subparsers = parser.add_subparsers(dest='command', help='Sub-commands')

    # run-all
    run_all_parser = subparsers.add_parser('run-all', help='Run all concepts')
    add_outdir_and_infiles(run_all_parser)
    add_run_all_args(run_all_parser)
    run_all_parser.add_argument('--package-name', required=True,
                                help='Name of package to run regular expressions from.')
    run_all_parser.add_argument('--include-text-output', action='store_true',
                                help='Include original text in output jsonl.')

    # run-all-matches
    run_all_matches_parser = subparsers.add_parser('run-all-matches', help='Run all concepts and output each match')
    add_outdir_and_infiles(run_all_matches_parser)
    add_run_all_args(run_all_matches_parser)
    run_all_matches_parser.add_argument('--package-name', required=True,
                                        help='Name of package to run regular expressions from.')

    # run4snippets
    run4snippets_parser = subparsers.add_parser('run4snippets', help='Extract snippets for review')
    add_outdir_and_infiles(run4snippets_parser)
    add_run_all_args(run4snippets_parser)
    run4snippets_parser.add_argument('--package-name', required=True,
                                     help='Name of package to run regular expressions from.')
    run4snippets_parser.add_argument('--context-length', type=int, default=180,
                                     help='Default context window.')
    run4snippets_parser.add_argument('--max-window', type=int, default=500,
                                     help='Maximum context window.')
    run4snippets_parser.add_argument('--target-concepts', nargs='+',
                                     help='Target concepts.')
    run4snippets_parser.add_argument('--target-categories', nargs='+',
                                     help='Target categories.')
    run4snippets_parser.add_argument('--order-metadata', nargs='+',
                                     help='Specify first metadata items in output.')

    # bio-tag
    bio_tag_parser = subparsers.add_parser('bio-tag', help='Generate BIO tagged data')
    add_outdir_and_infiles(bio_tag_parser)
    bio_tag_parser.add_argument('--package-name', required=True, help='Name of package.')

    # corpus2jsonl
    corpus2jsonl_parser = subparsers.add_parser('corpus2jsonl', help='Convert corpus to jsonl')
    add_outdir_and_infiles(corpus2jsonl_parser)
    corpus2jsonl_parser.add_argument('--split', choices=['sentence', 'chunk', 'window'], help='Split type')
    corpus2jsonl_parser.add_argument('--max-seq-length', dest='max_seq_length', default=512, type=int,
                                     help='Number of tokens to retain when `--split` is sent_chunk or sent_window.')
    corpus2jsonl_parser.add_argument('--tokenizer', default=None,
                                     help='Path to BERT tokenizer.')

    # create-bio-dataset
    create_bio_ds_parser = subparsers.add_parser('create-bio-dataset', help='Create BIO dataset for training')
    create_bio_ds_parser.add_argument('path', type=Path, help='Path to jsonl input file.')
    create_bio_ds_parser.add_argument('outpath', type=Path, help='Path to write created dataset.')
    create_bio_ds_parser.add_argument('--test-size', type=float, default=0.1)
    create_bio_ds_parser.add_argument('--validation-size', type=float, default=0.05)

    args = parser.parse_args()

    if not args.command:
        parser.print_help()
        sys.exit(1)

    cmd_args = vars(args)
    command = cmd_args.pop('command')

    # clean args (e.g. metadata_labels)
    cmd_args = clean_args(cmd_args)

    if command == 'run-all':
        run_all(**cmd_args)
    elif command == 'run-all-matches':
        run_all_matches(**cmd_args)
    elif command == 'run4snippets':
        run4snippets(**cmd_args)
    elif command == 'bio-tag':
        get_bio_tags(**cmd_args)
    elif command == 'corpus2jsonl':
        corpus2jsonl(**cmd_args)
    elif command == 'create-bio-dataset':
        create_bio_dataset(**cmd_args)


if __name__ == '__main__':
    main()



import re

import pytest

from konsepy.context.contexts import check_if_pattern_after, get_contexts


@pytest.mark.parametrize('pattern, word_window, exp_pre, exp_post, exp_around', [
    ('Väinämöinen', 2, 'The old ', ' played the', 'The old Väinämöinen played the'),
    ('kantele', 3, 'Väinämöinen played the ', '.', 'Väinämöinen played the kantele.'),
])
def test_get_contexts_word_window(pattern, word_window, exp_pre, exp_post, exp_around):
    text = 'The old Väinämöinen played the kantele.'
    m = re.search(pattern, text)
    res = get_contexts(m, text, word_window=word_window)
    assert res['precontext'] == exp_pre
    assert res['postcontext'] == exp_post
    assert res['around'] == exp_around


@pytest.mark.parametrize('pattern, pattern2, window, banned_characters, end, exp', [
    ('nose', 'help', 20, '.', None, None),
    ('nose', 'help', 20, '', None, 'help'),
])
def test_check_if_pattern_after(saiga_antelope, pattern, pattern2, window, banned_characters, end, exp):
    regex = re.compile(pattern2, re.I)
    if m := re.search(pattern, saiga_antelope, re.I):
        res = check_if_pattern_after(regex, m, saiga_antelope, window=window,
                                     banned_characters=banned_characters, end=end)
        if res:
            assert res.group() == exp
        else:
            assert res is None


@pytest.mark.parametrize('pattern, pattern2, window, banned_characters, end, exp', [
    ('help', 'nose', 20, '.', None, None),
    ('help', 'nose', 20, '', None, 'nose'),
])
def test_check_if_pattern_before(saiga_antelope, pattern, pattern2, window, banned_characters, end, exp):
    regex = re.compile(pattern2, re.I)
    if m := re.search(pattern, saiga_antelope, re.I):
        res = check_if_pattern_after(regex, m, saiga_antelope, window=window,
                                     banned_characters=banned_characters, end=end)
        if res:
            assert res.group() == exp
        else:
            assert res is None



import datetime
import re
from collections import Counter, defaultdict

from konsepy.constants import NOTEDATE_LABEL, ID_LABEL, NOTEID_LABEL, NOTETEXT_LABEL
from konsepy.context.contexts import get_contexts
from konsepy.importer import get_all_concepts
from konsepy.textio import iterate_csv_file, output_results
from loguru import logger

from konsepy.engine import ProcessingEngine


def run_regex_on_files(input_files, regex_func, package_name=None, *, start_after=0, stop_after=None,
                       require_regex=None, window_size=50,
                       id_label=ID_LABEL, noteid_label=NOTEID_LABEL,
                       notedate_label=NOTEDATE_LABEL, notetext_label=NOTETEXT_LABEL,
                       noteorder_label=None, metadata_labels=None,
                       select_probability=1.0, **kwargs):
    cat_counter_notes = Counter()
    cat_counter_mrns = defaultdict(set)
    noteid_to_cat = defaultdict(Counter)
    mrn_to_cat = defaultdict(Counter)
    unique_mrns = set()
    not_found_text = Counter()
    if require_regex:
        require_regex = re.compile(require_regex, re.I)

    count = 0
    for count, mrn, note_id, note_date, text, metadata in iterate_csv_file(
            input_files, start_after=start_after, stop_after=stop_after,
            id_label=id_label, noteid_label=noteid_label,
            notedate_label=notedate_label, notetext_label=notetext_label,
            noteorder_label=noteorder_label, metadata_labels=metadata_labels,
            select_probability=select_probability,
    ):
        if count % 50000 == 0:
            logger.info(
                f'Completed {count:,} records:'
                f' {len(unique_mrns):,} MRNs contain any category ({datetime.datetime.now()})')
        extract_categories(
            mrn, note_id, text, regex_func,
            cat_counter_mrns=cat_counter_mrns, cat_counter_notes=cat_counter_notes,
            mrn_to_cat=mrn_to_cat, require_regex=require_regex,
            not_found_text=not_found_text, noteid_to_cat=noteid_to_cat,
            unique_mrns=unique_mrns, window_size=window_size
        )
    logger.info(f'Finished. Total records: {count:,}  ({datetime.datetime.now()})')
    return cat_counter_notes, cat_counter_mrns, not_found_text, mrn_to_cat, noteid_to_cat


def extract_categories(mrn, note_id, text, regex_func, *, categories=None,
                       cat_counter_mrns=None, cat_counter_notes=None, mrn_to_cat=None,
                       not_found_text=None, noteid_to_cat=None,
                       require_regex=None, unique_mrns=None, window_size=50):
    if categories is None:  # don't re-run when empty list
        categories = list(regex_func(text, categories_only=True))
    for category in categories:
        mrn_to_cat[mrn][category] += 1
        noteid_to_cat[(mrn, note_id)][category] += 1
        cat_counter_notes[category] += 1
        cat_counter_mrns[category].add(mrn)
    if categories:
        unique_mrns.add(mrn)
    if not categories and not_found_text:
        if require_regex:
            for m in require_regex.finditer(text):
                start_snippet = max(0, m.start() - window_size)
                end_snippet = m.end() + window_size
                snippet = text[start_snippet:end_snippet + 1]
                not_found_text[' '.join(snippet.split())] += 1
        else:
            not_found_text[' '.join(text.split())] += 1


def run_regex_and_output(package_name, input_files, outdir, *concepts,
                         start_after=0, stop_after=None, require_regex=None, window_size=50,
                         id_label=ID_LABEL, noteid_label=NOTEID_LABEL,
                         notedate_label=NOTETEXT_LABEL, notetext_label=NOTETEXT_LABEL,
                         noteorder_label=None, select_probability=1.0, **kwargs):
    logger.info(f'Arguments ignored: {kwargs}')
    dt = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")

    # Use ProcessingEngine to consolidate concept loading and iteration
    engine = ProcessingEngine(
        input_files, package_name, start_after=start_after, stop_after=stop_after,
        id_label=id_label, noteid_label=noteid_label,
        notedate_label=notedate_label, notetext_label=notetext_label,
        noteorder_label=noteorder_label, select_probability=select_probability,
        concepts=concepts, **kwargs
    )

    # creates a separate output directory per concept
    for iconcept in engine.concepts:
        curr_outdir = outdir / f'{iconcept.name}_{dt}'
        curr_outdir.mkdir(parents=True)
        logger.add(curr_outdir / f'{iconcept.name}_{dt}.log')

        note_counter, cat_counter_mrns, not_found_text, mrn_to_cat, note_to_cat = run_regex_on_files(
            input_files, iconcept.run_func, package_name=package_name,
            start_after=start_after, stop_after=stop_after, require_regex=require_regex,
            window_size=window_size,
            id_label=id_label, noteid_label=noteid_label,
            notedate_label=notedate_label, notetext_label=notetext_label,
            noteorder_label=noteorder_label, select_probability=select_probability,
            **kwargs
        )
        output_results(curr_outdir, not_found_text=not_found_text, note_counter=note_counter,
                       cat_counter_mrns=cat_counter_mrns, category_enums=iconcept.category_enums,
                       note_to_cat=note_to_cat, mrn_to_cat=mrn_to_cat)



import argparse
import datetime
from pathlib import Path

from konsepy.constants import NOTETEXT_LABEL, NOTEDATE_LABEL, NOTEID_LABEL, ID_LABEL


def concept_cli(func):
    parser = argparse.ArgumentParser(fromfile_prefix_chars='@!')
    add_common_cli(parser)
    func(**clean_args(vars(parser.parse_args())))


def snippet_cli():
    parser = argparse.ArgumentParser(fromfile_prefix_chars='@!')
    parser.add_argument('--regexes', nargs='+',
                        help=r'REGEX_NAME==(?:re(?:gex)\sto\s(?:search|look)\sfor')
    parser.add_argument('--stop-after-regex-count', dest='stop_after_regex_count', default=None,
                        help='change to number if you want to limit number of regex "hits"; else keep None')
    parser.add_argument('--no-regex-func', dest='no_regex_func', action='store_true', default=False,
                        help='Run all regular expressions in `REGEXES` manaully, not calling `RUN_REGEX_FUNC`.')
    add_common_cli(parser)
    return clean_args(vars(parser.parse_args()))


def add_common_cli(parser: argparse.ArgumentParser):
    add_outdir_and_infiles(parser)
    parser.add_argument('--package-name', dest='package_name',
                        help='Name of package to run regular expressions from.')
    parser.add_argument('--concepts', nargs='+',
                        help='Name of concepts to process/run regular expressions for.')
    parser.add_argument('--require-regex', default=None,
                        help='Output text containing this regex but in which no regexes were found.')
    parser.add_argument('--start-after', default=0, type=int,
                        help='Start after skipping this many records')
    parser.add_argument('--stop-after', default=None, type=int,
                        help='change to number if you want to limit number of notes searched through; else None.')
    parser.add_argument('--select-probability', default=1.0, type=float,
                        help='Set to less than 1.0 to increase note sample (e.g., 0.3); 1.0=don\'t skip anything')
    parser.add_argument('--window-size', default=50, type=int,
                        help='Change the window for the pre/post contexts')


def add_outdir_and_infiles(parser: argparse.ArgumentParser = None):
    if not parser:
        parser = argparse.ArgumentParser(fromfile_prefix_chars='@!')
    parser.add_argument('--outdir', type=Path, default=Path('.'),
                        help='Directory to place output files.')
    parser.add_argument('--input-files', nargs='+', type=str, default=list(),
                        help='Input CSV, JSONL, or SAS file(s) to read.')
    parser.add_argument('--encoding', type=str, default='latin1',
                        help='Encoding for input files. Output files will be utf8.')
    parser.add_argument('--id-label', default=ID_LABEL,
                        help='Column label for individual id')
    parser.add_argument('--noteid-label', default=NOTEID_LABEL,
                        help='Column label for note id')
    parser.add_argument('--notedate-label', default=NOTEDATE_LABEL,
                        help='Column label for date of note (optional)')
    parser.add_argument('--notetext-label', default=NOTETEXT_LABEL,
                        help='Column label for note text')
    parser.add_argument('--noteorder-label', default=None,
                        help='Specify column that enumerates the order of multiple parts of a single note. (optional)')
    parser.add_argument('--metadata-labels', default=None, nargs='*',
                        help='Specify additional column metadata as kwarg_name==in_dataset_name')
    parser.add_argument('--limit-noteids', default=None, nargs='+',
                        help='Limit run to only this noteid (useful for debugging).')
    return parser


def add_run_all_args(parser: argparse.ArgumentParser = None):
    if not parser:
        parser = argparse.ArgumentParser(fromfile_prefix_chars='@!')
    parser.add_argument('--incremental-output-only', action='store_true', default=False,
                        help='Do not retain summarized output, just output incremental jsonl file.')
    parser.add_argument('--concepts', nargs='+', default=False,
                        help='Look for these particular concepts.')
    parser.add_argument('--window-size', dest='window', default=30, type=int,
                        help='Change the window for the pre/post contexts')
    parser.add_argument('--word-window', dest='word_window', default=None, type=int,
                        help='Change the word window for the pre/post contexts')


def _get_casting_func(target, format_=None):
    match target:
        case 'dt':
            if format_:
                return lambda x: datetime.datetime.strptime(x, format_)
            else:
                raise ValueError(f'Parsing datetime requires format string e.g., {target}==%m/%d/%y.'
                                 f' For info on formatting,'
                                 f' see https://docs.python.org/3/library/datetime.html#strftime-strptime-behavior')
        case 'int':
            return lambda x: int(x)
        case 'float':
            return lambda x: float(x)
        case 'str':
            return lambda x: str(x)
    raise ValueError(f'Unrecognized cast request: {target}=={format_}')


def clean_metadata_labels(metadata_labels):
    result = {}
    for md_labels in metadata_labels:
        if '==' in md_labels:
            src, dest, *other = md_labels.split('==')
        else:
            src = md_labels
            dest = md_labels
            other = None
        func = _get_casting_func(*other) if other else lambda x: x
        result[src] = (dest, func)
    return result


def clean_args(args: dict):
    """Fix/format arguments, e.g., `metadata-labels`"""
    if args.get('metadata_labels', None):
        args['metadata_labels'] = clean_metadata_labels(args['metadata_labels'])
    return args


def parse_and_clean_args(parser: argparse.ArgumentParser) -> dict:
    """Wrapper to parse an clean arguments."""
    return clean_args(vars(parser.parse_args()))



import re

from konsepy.context.contexts import get_contexts

OTHER_SUBJECT = [
    'brother', 'sister', 'sibling', 'mother', 'mom', 'father', 'dad',
    'parent', 'dtr', 'daughter', 'son', 'child', 'children',
    'aunt', 'uncle', 'niece', 'nephew', 'cousin',
    'gma', 'grandma', 'grandmother', 'gpa', 'grandpa', 'grandfather',
    'grandparent', 'grandparents', 'grand\W*parent\(s\)',
    r'grand\W?daughter', 'gdtr',
    'husb(?:and)?', 'wife', 'wives', 'spouse', 'partner', r'(?:girl|boy)\W?friend',
    'neighbou?r', r'significant\W*other', 'person',
    'people', 'friend', 'kid', 'peer',
    'physician', 'doctor', 'family', r'co\W?worker',
    r'(?:house|room|flat)\W?mate', 'colleague', 'employe[re]', 'others',
    'family (?:history|hx)', 'maternal', 'paternal',
    'parents', 'parent\(s\)',
]

OTHER_SUBJECT_RX = re.compile('(?:{})'.format(
    '|'.join(rf'\b{x}\'?\b' for x in OTHER_SUBJECT))
)


def has_other_subject(text, direction=0, other_subject_rx=OTHER_SUBJECT_RX, banned_characters='.'):
    """Return first mention of other subject

    direction = set to -1 (precontext) or 1 (postcontext) to ensure is in same sentence
    """
    if m := other_subject_rx.search(text):
        if direction == -1:  # precontext
            for ch in banned_characters:
                if ch in text[m.end():]:
                    return None
        elif direction == 1:  # postcontext
            for ch in banned_characters:
                if ch in text[:m.start()]:
                    return None
        return m
    return None


# per/according to/at X's house
PER_PAT = re.compile(r'(?:\bper\b|according\W*to|\bat\b|\bdue\W*to\b|because\W*of\b)(?:\W+\w+)?\W*$', re.I)
OBJECT_PAT = re.compile(r'(?:\w+\W+)?(?:med\w*|gun|weapon|pill)s?', re.I)


def is_not_other_subject(m, precontext, postcontext, **kwargs):
    """Exceptions to other subject pattern"""
    if PER_PAT.search(precontext) or OBJECT_PAT.search(postcontext):
        return True
    return False


def check_if_other_subject(m, precontext, postcontext, text, window=30, banned_characters='.',
                           other_concept=True, **kwargs):
    if m2 := has_other_subject(precontext, direction=-1, banned_characters=banned_characters):
        if is_not_other_subject(**get_contexts(
                m2, text, context_match=m, context_window=window, context_direction=-1,
        )):
            pass  # might still be in post-context
        else:
            return m2 if other_concept is True else other_concept
    if m2 := has_other_subject(postcontext, direction=1, banned_characters=banned_characters):
        if is_not_other_subject(**get_contexts(
                m2, text, context_match=m, context_window=window, context_direction=1,
        )):
            pass
        else:
            return m2 if other_concept is True else other_concept
    return None



import datetime
import json
import pathlib
from collections import Counter, defaultdict

from loguru import logger

from konsepy.cli import add_outdir_and_infiles, add_run_all_args, clean_args
from konsepy.constants import NOTEDATE_LABEL, ID_LABEL, NOTEID_LABEL, NOTETEXT_LABEL
from konsepy.importer import get_all_concepts
from konsepy.regex import extract_categories
from konsepy.textio import iterate_csv_file, output_results

from konsepy.engine import ProcessingEngine


def run_all(input_files, outdir: pathlib.Path, package_name: str, *,
            encoding='latin1', id_label=ID_LABEL, noteid_label=NOTEID_LABEL,
            notedate_label=NOTEDATE_LABEL, notetext_label=NOTETEXT_LABEL,
            noteorder_label=None, metadata_labels=None, incremental_output_only=False,
            concepts=None, include_text_output=False, limit_noteids=None, **kwargs) -> pathlib.Path:
    """
    Run all concepts.
    Return: Newly created `run_all` directory.
    """
    logger.info(f'Arguments ignored: {kwargs}')
    dt = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
    curr_outdir = outdir / f'run_all_{dt}'
    curr_outdir.mkdir(parents=True)
    logger.add(curr_outdir / f'run_all_{dt}.log')

    cat_counter_notes = Counter()
    cat_counter_mrns = defaultdict(set)
    noteid_to_cat = defaultdict(Counter)
    mrn_to_cat = defaultdict(Counter)
    unique_mrns = set()

    engine = ProcessingEngine(
        input_files, package_name, encoding=encoding, id_label=id_label,
        noteid_label=noteid_label, notedate_label=notedate_label,
        notetext_label=notetext_label, noteorder_label=noteorder_label,
        metadata_labels=metadata_labels, concepts=concepts,
        limit_noteids=limit_noteids, **kwargs
    )

    with open(curr_outdir / 'output.jsonl', 'w') as out:
        def callback(studyid, note_id, note_date, text, metadata, concept, categories, matches):
            if categories:
                out.write(json.dumps({
                    'studyid': studyid,
                    'note_id': note_id,
                    'note_date': note_date,
                    'text': text if include_text_output else None,
                    'concept': concept.name,
                    'matches': [m.group() for m in matches] if matches else None,
                    'categories': categories,
                }) + '\n')
            if not incremental_output_only:
                extract_categories(
                    studyid, note_id, text, concept.run_func, categories=categories,
                    cat_counter_mrns=cat_counter_mrns, cat_counter_notes=cat_counter_notes,
                    mrn_to_cat=mrn_to_cat, noteid_to_cat=noteid_to_cat,
                    unique_mrns=unique_mrns
                )

        engine.run(callback)

    if not incremental_output_only:
        logger.info(f'Bulk writing to {curr_outdir}.')
        output_results(curr_outdir, note_counter=cat_counter_notes,
                       cat_counter_mrns=cat_counter_mrns,
                       category_enums=[category_enum for c in engine.concepts for category_enum in c.category_enums],
                       note_to_cat=noteid_to_cat, mrn_to_cat=mrn_to_cat)
    return curr_outdir


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(fromfile_prefix_chars='@!')
    add_outdir_and_infiles(parser)
    add_run_all_args(parser)
    run_all(**clean_args(vars(parser.parse_args())))



import re
from enum import Enum
from konsepy.rxsearch import (
    search_first_regex,
    search_all_regex,
    get_all_regex_by_index,
    search_all_regex_match_func,
    search_and_replace_regex_func,
    search_first_regex_func,
    search_all_regex_func,
)


class KalevalaCategory(Enum):
    HERO = 1
    GOD = 2
    SMITH = 3
    UNKNOWN = 4


def test_search_first_regex():
    regexes = [
        (re.compile(r'Väinämöinen', re.I), KalevalaCategory.HERO),
        (re.compile(r'Ilmarinen', re.I), KalevalaCategory.SMITH),
    ]
    func = search_first_regex(regexes)

    text = 'Väinämöinen and Ilmarinen were friends.'
    results = list(func(text))
    assert results == [KalevalaCategory.HERO]

    text2 = 'Ilmarinen was a smith.'
    results2 = list(func(text2))
    assert results2 == [KalevalaCategory.SMITH]


def test_search_all_regex():
    regexes = [
        (re.compile(r'Väinämöinen', re.I), KalevalaCategory.HERO),
        (re.compile(r'Ilmarinen', re.I), KalevalaCategory.SMITH),
    ]
    func = search_all_regex(regexes)

    text = 'Väinämöinen and Ilmarinen were friends. Väinämöinen sang.'
    results = list(func(text))
    # search_all_regex iterates over regexes, then finditer for each
    # So it returns all matches for HERO first, then all for SMITH
    assert results == [KalevalaCategory.HERO, KalevalaCategory.HERO, KalevalaCategory.SMITH]


def test_get_all_regex_by_index():
    regexes = [
        (re.compile(r'Ukko', re.I), KalevalaCategory.GOD),
    ]
    func = get_all_regex_by_index(regexes)

    text = 'Ukko is the god of sky.'
    results = list(func(text))
    assert len(results) == 1
    assert results[0][0] == KalevalaCategory.GOD
    assert results[0][1] == 'Ukko'
    assert results[0][2] == 0
    assert results[0][3] == 4


def test_search_all_regex_match_func():
    def check_context(**kwargs):
        if 'old' in kwargs.get('text', '').lower():
            return KalevalaCategory.HERO
        return None

    regexes = [
        (re.compile(r'Väinämöinen', re.I), KalevalaCategory.UNKNOWN, check_context),
    ]
    func = search_all_regex_match_func(regexes)

    text = 'Old Väinämöinen sang.'
    results = list(func(text))
    assert results == [KalevalaCategory.HERO]

    text2 = 'Young Väinämöinen.'
    results2 = list(func(text2))
    # if func returns None, it falls back to category
    assert results2 == [KalevalaCategory.UNKNOWN]


def test_search_all_regex_match_func2():
    def check_context(precontext, **kwargs):
        if 'old' in precontext.lower():
            return KalevalaCategory.HERO
        return None

    regexes = [
        (re.compile(r'Väinämöinen', re.I), KalevalaCategory.UNKNOWN, check_context),
    ]
    func = search_all_regex_match_func(regexes)

    text = 'Old Väinämöinen sang.'
    results = list(func(text))
    assert results == [KalevalaCategory.HERO]

    text2 = 'Young Väinämöinen.'
    results2 = list(func(text2))
    # if func returns None, it falls back to category
    assert results2 == [KalevalaCategory.UNKNOWN]


def test_search_and_replace_regex_func():
    regexes = [
        (re.compile(r'Väinämöinen', re.I), KalevalaCategory.HERO),
        (re.compile(r'Väinä', re.I), KalevalaCategory.UNKNOWN),
    ]
    func = search_and_replace_regex_func(regexes)

    text = 'Väinämöinen'
    results = list(func(text))
    # first regex matches Väinämöinen, then it's replaced by dots, so 'Väinä' won't match anymore
    assert results == [KalevalaCategory.HERO]


def test_search_first_regex_func():
    regexes = [
        (re.compile(r'Väinämöinen', re.I), KalevalaCategory.HERO),
        (re.compile(r'Ilmarinen', re.I), KalevalaCategory.SMITH),
    ]
    func = search_first_regex_func(regexes)

    text = 'Väinämöinen and Ilmarinen.'
    results = list(func(text))
    assert results == [KalevalaCategory.HERO]


def test_search_all_regex_func_sentinel():
    regexes = [
        (re.compile(r'Väinämöinen', re.I), KalevalaCategory.HERO),
        (None, None),  # sentinel
        (re.compile(r'Ukko', re.I), KalevalaCategory.GOD),
    ]
    func = search_all_regex_func(regexes)

    text = 'Väinämöinen and Ukko.'
    results = list(func(text))
    # found HERO, so it stops at sentinel.
    assert results == [KalevalaCategory.HERO]

    text2 = 'Ukko alone.'
    results2 = list(func(text2))
    # HERO not found, so it continues past sentinel.
    assert results2 == [KalevalaCategory.GOD]


def test_search_all_regex_func_with_indices():
    def find_indices(text):
        yield 0, 10

    regexes = [
        (re.compile(r'Ukko', re.I), KalevalaCategory.GOD, [None], find_indices),
    ]
    func = search_all_regex_func(regexes)

    text = 'Ukko is here. Ukko is there.'
    results = list(func(text))
    # Only the first Ukko is in the first 10 chars.
    assert results == [KalevalaCategory.GOD]


def test_search_all_regex_func_word_window():
    def check_context(precontext, **kwargs):
        # With word_window=2, precontext for 'sang' should be 'Old Väinämöinen '
        if 'Old Väinämöinen' in precontext:
            return 'SINGER'

    regexes = [
        (re.compile(r'sang', re.I), 'UNKNOWN', check_context),
    ]
    # Use word_window=2
    func = search_all_regex_func(regexes, word_window=2)

    text = 'The wise Old Väinämöinen sang a song.'
    results = list(func(text))
    assert results == ['SINGER']

    # If word_window=1, precontext would be 'Väinämöinen '
    func2 = search_all_regex_func(regexes, word_window=1)
    results2 = list(func2(text))
    assert results2 == ['UNKNOWN']


def test_search_and_replace_regex_func_word_window():
    def check_context(postcontext, **kwargs):
        # with word_window=2, postcontext for 'Ilmarinen' should be ' forged the'
        if 'forged' in postcontext:
            return 'BLACKSMITH'

    regexes = [
        (re.compile(r'Ilmarinen', re.I), 'UNKNOWN', check_context),
    ]
    func = search_and_replace_regex_func(regexes, word_window=2)

    text = 'Ilmarinen forged the Sampo.'
    results = list(func(text))
    assert results == ['BLACKSMITH']


def test_search_first_regex_func_word_window():
    def check_context(around, **kwargs):
        # with word_window=1, around for 'Sampo' should be 'forged Sampo.'
        if 'forged' in around:
            return 'ARTEFACT'

    regexes = [
        (re.compile(r'Sampo', re.I), 'UNKNOWN', check_context),
    ]
    func = search_first_regex_func(regexes, word_window=1)

    text = 'Ilmarinen forged Sampo.'
    results = list(func(text))
    assert results == ['ARTEFACT']



import importlib
import inspect
import pkgutil
from enum import EnumMeta
from pathlib import Path

from loguru import logger


class ConceptImport:

    def __init__(self, module_info, package_name):
        self.name = module_info.name
        self.imp = importlib.import_module(f'{package_name}.concepts.{self.name}')
        self.category_enums = list(self._get_categories())
        self._run_func = self.imp.RUN_REGEXES_FUNC
        self.regexes = self.imp.REGEXES
        self._params = inspect.signature(self._run_func).parameters

        self.has_include_match = self.has_param('include_match')

    def has_param(self, param, default=False):
        exists = param in self._params
        if not exists:
            logger.warning(f'Concept `{self.name}` is missing `{param}={default}` in RUN_REGEXES_FUNC.')
        return exists

    def run_func(self, text, include_match=True, categories_only=False, **metadata):
        # ensure requested metadata in calling function
        metadata = {k: v for k, v in metadata.items() if k in self._params}
        if include_match and self.has_include_match:
            res = list(self._run_func(text, include_match=include_match, **metadata))
            # matches = [(m.group(), m.start(), m.end()) for _, m in res]
            matches = [m for _, m in res]  # return match directly
            categories = [str(category) for category, _ in res]
        else:
            if self.has_include_match:
                categories = [str(category) for category in self._run_func(text, include_match=False, **metadata)]
            else:
                categories = [str(category) for category in self._run_func(text, **metadata)]
            matches = None
        if categories_only:
            return categories
        return categories, matches

    def run(self, sentence):
        return self.run_func(sentence)

    @property
    def domain(self):
        return self.name

    @property
    def categories(self):
        return [category.name for category_enum in self.category_enums for category in category_enum]

    def _get_categories(self):
        categories = []
        for name, value in self.imp.__dict__.items():
            if isinstance(value, EnumMeta):
                categories.append(value)
        if categories:
            return categories
        else:
            raise ValueError(f'Unable to identify category enum for concept "{self.name}".')

    def __str__(self):
        return f'ConceptImport<{self.name}>'


def get_all_concepts(package_name: str, *concepts):
    imp = importlib.import_module(f'{package_name}.concepts')
    path = Path(imp.__file__).parent
    for module_info in pkgutil.iter_modules([path]):
        if concepts and module_info.name not in concepts:
            continue  # look for only requested concepts if any supplied
        try:
            yield ConceptImport(module_info, package_name)
        except ValueError as ve:
            logger.warning(f'Failed to load concept: {package_name}.concepts.{module_info.name}')
            logger.exception(ve)



import json
import sys
from pathlib import Path
from unittest.mock import patch

import pytest
from konsepy.main import main


try:
    import spacy
    HAS_SPACY = True
except ImportError:
    HAS_SPACY = False

try:
    import datasets
    HAS_DATASETS = True
except ImportError:
    HAS_DATASETS = False


@pytest.fixture
def input_file(request):
    return Path(request.fspath).parent / 'data' / 'corpus.jsonl'


def test_cli_run_all(tmp_path, input_file):
    outdir = tmp_path / 'run_all_out'
    outdir.mkdir()

    test_args = [
        'konsepy',
        'run-all',
        '--input-files', str(input_file),
        '--outdir', str(outdir),
        '--package-name', 'example_nlp',
        '--id-label', 'chapter',
        '--noteid-label', 'chapter',
    ]

    with patch.object(sys, 'argv', test_args):
        main()

    # verify that an output directory was created (it has a timestamp)
    run_all_dirs = list(outdir.glob('run_all_*'))
    assert len(run_all_dirs) == 1
    actual_outdir = run_all_dirs[0]

    assert (actual_outdir / 'output.jsonl').exists()
    assert (actual_outdir / 'category_counts.csv').exists()

    with open(actual_outdir / 'output.jsonl', encoding='utf8') as f:
        lines = f.readlines()
        assert len(lines) > 0
        first_entry = json.loads(lines[0])
        assert 'concept' in first_entry
        assert 'categories' in first_entry


def test_cli_run_all_matches(tmp_path, input_file):
    outdir = tmp_path / 'run_all_matches_out'
    outdir.mkdir()

    test_args = [
        'konsepy',
        'run-all-matches',
        '--input-files', str(input_file),
        '--outdir', str(outdir),
        '--package-name', 'example_nlp',
        '--id-label', 'chapter',
        '--noteid-label', 'chapter',
    ]

    with patch.object(sys, 'argv', test_args):
        main()

    # verify that an output directory was created (it has a timestamp)
    run_all_dirs = list(outdir.glob('run_all_matches_*'))
    assert len(run_all_dirs) == 1
    actual_outdir = run_all_dirs[0]

    assert (actual_outdir / 'output.jsonl').exists()

    with open(actual_outdir / 'output.jsonl', encoding='utf8') as f:
        lines = f.readlines()
        assert len(lines) > 0
        first_entry = json.loads(lines[0])
        assert 'concept' in first_entry
        assert 'category' in first_entry
        assert 'precontext' in first_entry
        assert 'match' in first_entry
        assert 'postcontext' in first_entry
        assert 'start_index' in first_entry
        assert 'end_index' in first_entry


def test_cli_run4snippets(tmp_path, input_file):
    outdir = tmp_path / 'snippets_out'
    outdir.mkdir()

    test_args = [
        'konsepy',
        'run4snippets',
        '--input-files', str(input_file),
        '--outdir', str(outdir),
        '--package-name', 'example_nlp',
        '--id-label', 'chapter',
        '--noteid-label', 'chapter',
    ]

    with patch.object(sys, 'argv', test_args):
        main()

    snippet_dirs = list(outdir.glob('run_all_*'))
    assert len(snippet_dirs) == 1
    actual_outdir = snippet_dirs[0]

    assert (actual_outdir / 'output.jsonl').exists()
    with open(actual_outdir / 'output.jsonl', encoding='utf8') as f:
        lines = f.readlines()
        assert len(lines) > 0
        first_entry = json.loads(lines[0])
        assert 'precontext' in first_entry
        assert 'postcontext' in first_entry
        assert 'match' in first_entry


def test_cli_bio_tag(tmp_path, input_file):
    outdir = tmp_path / 'bio_tag_out'
    outdir.mkdir()

    test_args = [
        'konsepy',
        'bio-tag',
        '--input-files', str(input_file),
        '--outdir', str(outdir),
        '--package-name', 'example_nlp',
        '--id-label', 'chapter',
        '--noteid-label', 'chapter',
    ]

    with patch.object(sys, 'argv', test_args):
        main()

    assert (outdir / 'bio_tag_data.csv').exists()
    assert (outdir / 'bio_tag_data.jsonl').exists()


@pytest.mark.skipif(not HAS_SPACY, reason='spacy not installed')
def test_cli_corpus2jsonl(tmp_path, input_file):
    outdir = tmp_path / 'corpus2jsonl_out'
    outdir.mkdir()

    test_args = [
        'konsepy',
        'corpus2jsonl',
        '--input-files', str(input_file),
        '--outdir', str(outdir),
        '--id-label', 'chapter',
        '--noteid-label', 'chapter',
    ]

    with patch.object(sys, 'argv', test_args):
        main()

    # corpus2jsonl might output to outdir directly or a subdirectory
    jsonl_files = list(outdir.glob('*.jsonl'))
    assert len(jsonl_files) >= 1


@pytest.mark.skipif(not HAS_DATASETS, reason='datasets not installed')
def test_cli_create_bio_dataset(tmp_path, input_file):
    # we need a bio_tag_data.jsonl first.
    bio_out = tmp_path / 'bio_tag_for_ds'
    bio_out.mkdir()
    
    with patch.object(sys, 'argv', [
        'konsepy', 'bio-tag', '--input-files', str(input_file),
        '--outdir', str(bio_out), '--package-name', 'example_nlp',
        '--id-label', 'chapter', '--noteid-label', 'chapter'
    ]):
        main()
    
    bio_jsonl = bio_out / 'bio_tag_data.jsonl'
    ds_out = tmp_path / 'dataset_out'
    ds_out.mkdir()

    test_args = [
        'konsepy',
        'create-bio-dataset',
        str(bio_jsonl),
        str(ds_out),
    ]

    with patch.object(sys, 'argv', test_args):
        main()

    # datasets.save_to_disk creates a directory
    ds_dirs = list(ds_out.glob('*.dataset'))
    assert len(ds_dirs) == 1
    assert ds_dirs[0].is_dir()



import csv
import datetime
import itertools
import re

from loguru import logger

from konsepy.cli import snippet_cli
from konsepy.constants import NOTEDATE_LABEL, ID_LABEL, NOTEID_LABEL, NOTETEXT_LABEL
from konsepy.importer import get_all_concepts
from konsepy.textio import iterate_csv_file


def get_text_snippets_regexes(input_files, outdir, regexes, *, start_after=0, stop_after=None, window_size=50,
                              encoding='latin1', id_label=ID_LABEL, noteid_label=NOTEID_LABEL,
                              notedate_label=NOTEDATE_LABEL, notetext_label=NOTETEXT_LABEL,
                              noteorder_label=None, metadata_labels=None,
                              select_probability=1.0, label='snippets', stop_after_regex_count=None, **kwargs):
    dt = datetime.datetime.now().strftime('%Y%m%d_%H%M%S')
    logger.warning('Snippets will have spaces normalized:'
                   ' multiple spaces/newlines/tabs will be converted'
                   ' to a single space in the output.')
    rx_count = 0
    outdir.mkdir(exist_ok=True)
    outfile = outdir / f'{label}_{dt}.csv'
    with open(outfile, 'w', newline='') as out:
        writer = csv.writer(out)
        writer.writerow(['id', 'studyid', 'note_id', 'date', 'regex_name', 'precontext', 'term', 'postcontext'])
        for i, (_, studyid, note_id, note_date, text, metadata) in enumerate(iterate_csv_file(
                input_files, encoding=encoding,
                start_after=start_after, stop_after=stop_after,
                id_label=id_label, noteid_label=noteid_label,
                notetext_label=notetext_label, notedate_label=notedate_label,
                noteorder_label=noteorder_label, metadata_labels=metadata_labels,
                select_probability=select_probability,
        ), start=1):
            text = ' '.join(text.split())  # remove newlines, etc. (bad for snippets in Excel)
            for regex_ in regexes:
                if isinstance(regex_, (str, tuple)):
                    if isinstance(regex_, str):
                        name, regex = regex_.split('==')
                    elif isinstance(regex_, tuple):
                        name, regex = regex_
                    if isinstance(regex, str):
                        regex = re.compile(regex, re.I)
                    func = lambda x: zip(itertools.repeat(name), regex.finditer(x))
                elif callable(regex_):
                    func = lambda x: regex_(x, include_match=True, **metadata)
                else:
                    raise ValueError(f'Unknown how to handle regular expression of type {type(regex_)}: {regex_}')
                for name, m in zip(func(text)):
                    precontext = text[max(m.start() - window_size, 0):m.start()]
                    postcontext = text[m.end():m.end() + window_size]
                    writer.writerow([
                        rx_count,  # id
                        studyid,
                        note_id,
                        note_date,
                        name,
                        precontext,
                        m.group(),  # term
                        postcontext,
                    ])
                    rx_count += 1
                    if stop_after_regex_count and rx_count >= stop_after_regex_count:
                        return
            if i % 1000 == 0:
                logger.info(f'Completed {i} records, found {rx_count} matches.')
    return outfile


def get_text_snippets_for_concept_algorithm(package_name, input_files, outdir, *, concepts=None,
                                            encoding='latin1',
                                            start_after=0, stop_after=None, window_size=50,
                                            id_label=ID_LABEL, noteid_label=NOTEID_LABEL,
                                            notedate_label=NOTEDATE_LABEL, notetext_label=NOTETEXT_LABEL,
                                            noteorder_label=None,
                                            select_probability=1.0, label='snippets', stop_after_regex_count=None,
                                            no_regex_func=False,
                                            **kwargs):
    if concepts is None:
        concepts = list()
    if no_regex_func:  # ignore negation, add functions, the RUN_REGEX_FUNC
        regexes = [(category, regex)
                   for concept in get_all_concepts(package_name, *concepts)
                   for regex, category, *_ in concept.regexes]
    else:
        regexes = [concept.run_func for concept in get_all_concepts(package_name, *concepts)]

    return get_text_snippets_regexes(input_files, outdir, regexes,
                                     encoding=encoding,
                                     start_after=start_after,
                                     stop_after=stop_after,
                                     window_size=window_size,
                                     select_probability=select_probability,
                                     id_label=id_label,
                                     noteid_label=noteid_label,
                                     notedate_label=notedate_label,
                                     notetext_label=notetext_label,
                                     noteorder_label=noteorder_label,
                                     label=label,
                                     stop_after_regex_count=stop_after_regex_count,
                                     )


def get_text_snippets_cli(package_name=None):
    kwargs = snippet_cli()
    kwargs['package_name'] = kwargs.get('package_name', package_name) or package_name

    if package_name and kwargs.get('regexes', None) is None:
        return get_text_snippets_for_concept_algorithm(**kwargs)
    else:
        return get_text_snippets_regexes(**kwargs)


if __name__ == '__main__':
    get_text_snippets_cli()



import datetime
import json
import pathlib
from loguru import logger

from konsepy.constants import NOTEDATE_LABEL, ID_LABEL, NOTEID_LABEL, NOTETEXT_LABEL
from konsepy.context.contexts import get_contexts
from konsepy.engine import ProcessingEngine


def run_all_matches(input_files, outdir: pathlib.Path, package_name: str, *,
                    encoding='latin1', id_label=ID_LABEL, noteid_label=NOTEID_LABEL,
                    notedate_label=NOTEDATE_LABEL, notetext_label=NOTETEXT_LABEL,
                    noteorder_label=None, metadata_labels=None,
                    concepts=None, limit_noteids=None, window=30, word_window=None, **kwargs) -> pathlib.Path:
    """
    Run all concepts and output each match as a separate JSONL row.
    Return: Newly created `run_all_matches` directory.
    """
    if kwargs:
        logger.info(f'Arguments ignored: {kwargs}')
    dt = datetime.datetime.now().strftime('%Y%m%d_%H%M%S')
    curr_outdir = outdir / f'run_all_matches_{dt}'
    curr_outdir.mkdir(parents=True)
    logger.add(curr_outdir / f'run_all_matches_{dt}.log')

    engine = ProcessingEngine(
        input_files, package_name, encoding=encoding, id_label=id_label,
        noteid_label=noteid_label, notedate_label=notedate_label,
        notetext_label=notetext_label, noteorder_label=noteorder_label,
        metadata_labels=metadata_labels, concepts=concepts,
        limit_noteids=limit_noteids, **kwargs
    )

    output_path = curr_outdir / 'output.jsonl'
    with open(output_path, 'w', encoding='utf8') as out:
        def callback(studyid, note_id, note_date, text, metadata, concept, categories, matches):
            if not matches:
                return

            for category, m in zip(categories, matches):
                # serialize category cleanly
                cat_name = category.name if hasattr(category, 'name') else str(category)

                contexts = get_contexts(m, text, window=window, word_window=word_window)
                row = {
                    'studyid': studyid,
                    'note_id': note_id,
                    'note_date': note_date,
                    'concept': concept.name,
                    'category': cat_name,
                    'precontext': contexts['precontext'],
                    'match': m.group(),
                    'postcontext': contexts['postcontext'],
                    'start_index': m.start(),
                    'end_index': m.end(),
                }
                if metadata:
                    row.update(metadata)
                if group_dict := m.groupdict():
                    row.update(group_dict)

                out.write(json.dumps(row) + '\n')

        engine.run(callback)

    return curr_outdir



import datetime
import json
import pathlib

from loguru import logger

from konsepy.cli import add_outdir_and_infiles, add_run_all_args, clean_args, clean_metadata_labels
from konsepy.constants import NOTEDATE_LABEL, ID_LABEL, NOTEID_LABEL, NOTETEXT_LABEL
from konsepy.engine import ProcessingEngine


def _retain_record(concept, category, target_categories, target_concepts):
    if not target_concepts and not target_categories:
        return True
    if target_categories and category in target_categories:
        return True
    if target_concepts and concept.name in target_concepts:
        return True
    return False


def run4snippets(input_files, outdir: pathlib.Path, package_name: str, *,
                 context_length=180, max_window=500,
                 target_categories=None, target_concepts=None,
                 order_metadata=None,
                 encoding='utf8', id_label=ID_LABEL, noteid_label=NOTEID_LABEL,
                 notedate_label=NOTEDATE_LABEL, notetext_label=NOTETEXT_LABEL,
                 noteorder_label=None, metadata_labels=None,
                 concepts=None, limit_noteids=None, **kwargs) -> pathlib.Path:
    """
    Run all concepts.
    Return: Newly created `run_all` directory.
    """
    logger.info(f'Arguments ignored: {kwargs}')
    target_categories = set(target_categories) if target_categories else set()
    target_concepts = set(target_concepts) if target_concepts else set()
    if isinstance(metadata_labels, list):
        metadata_labels = clean_metadata_labels(metadata_labels or dict())
    elif metadata_labels is None:
        metadata_labels = dict()
    order_metadata = order_metadata or list()

    dt = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
    curr_outdir = outdir / f'run_all_{dt}'
    curr_outdir.mkdir(parents=True)
    logger.add(curr_outdir / f'run_all_{dt}.log')

    output_length = [0]  # Use list to allow modification in callback

    engine = ProcessingEngine(
        input_files, package_name, encoding=encoding, id_label=id_label,
        noteid_label=noteid_label, notedate_label=notedate_label,
        notetext_label=notetext_label, noteorder_label=noteorder_label,
        metadata_labels=metadata_labels, concepts=concepts,
        limit_noteids=limit_noteids, **kwargs
    )

    all_keys = [
                   'note_id', 'concept', 'category', 'studyid', 'note_date', 'match',
               ] + list(metadata_labels.keys()) + [
                   'start_index', 'end_index', 'precontext', 'postcontext', 'pretext', 'posttext',
               ]
    ordered_keys = order_metadata + [key for key in all_keys if key not in order_metadata]

    with open(curr_outdir / 'output.jsonl', 'w') as out:
        def callback(studyid, note_id, note_date, text, metadata, concept, categories, matches):
            for m, category in zip(matches, categories):
                if _retain_record(concept, category, target_categories, target_concepts):
                    output_length[0] += 1
                    curr_data = {
                                    'note_id': note_id,
                                    'concept': concept.name,
                                    'category': category,
                                    'studyid': studyid,
                                    'note_date': note_date,
                                    'match': m.group(),
                                    'start_index': m.start(),
                                    'end_index': m.end(),
                                    'precontext': text[max(m.start() - context_length, 0): m.start()],
                                    'postcontext': text[m.end(): m.end() + context_length],
                                    'pretext': text[max(m.start() - max_window, 0): m.start()],
                                    'posttext': text[m.end(): m.end() + max_window],
                                } | metadata
                    out.write(json.dumps({k: curr_data[k] for k in ordered_keys}) + '\n')

        engine.run(callback)

    target_concepts_str = '", "'.join(target_concepts) if target_concepts else 'all'
    target_categories_str = '", "'.join(target_categories) if target_categories else 'all'
    logger.info(
        f'Output {output_length[0]:,} rows belonging to'
        f' concepts: "{target_concepts_str}";'
        f' and categories: "{target_categories_str}".')
    return curr_outdir


def main(package_name=None):
    import argparse

    parser = argparse.ArgumentParser(fromfile_prefix_chars='@!')
    add_outdir_and_infiles(parser)
    add_run_all_args(parser)
    parser.add_argument('--context-length', dest='context_length', type=int, default=180,
                        help='Default context window to show around match.')
    parser.add_argument('--max-window', dest='max_window', type=int, default=500,
                        help='Maximum context window that will be output and available for review.')
    parser.add_argument('--target-concepts', dest='target_concepts', nargs='+',
                        help='Target concepts (exclude others) of form `jealousy`.'
                             ' Treated as "or" when combined with `--target-categories`.')
    parser.add_argument('--target-categories', dest='target_categories', nargs='+',
                        help='Target categories (exclude others) of form `Jealous.YES`.'
                             ' Treated as "or" when combined with `--target-concepts`.')
    parser.add_argument('--order-metadata', dest='order_metadata', nargs='+',
                        help='Specify the first metadata items to appear in jsonlines output.'
                             ' E.g., textual_review_app will display on the first 3 metadata items.')
    if package_name:
        run4snippets(package_name=package_name, **clean_args(vars(parser.parse_args())))
    else:
        run4snippets(**clean_args(vars(parser.parse_args())))


if __name__ == '__main__':
    main()



import datetime
from loguru import logger
from konsepy.importer import get_all_concepts
from konsepy.textio import iterate_csv_file
from konsepy.constants import NOTEDATE_LABEL, ID_LABEL, NOTEID_LABEL, NOTETEXT_LABEL

class ProcessingEngine:
    def __init__(self, input_files, package_name, *,
                 encoding='latin1', id_label=ID_LABEL, noteid_label=NOTEID_LABEL,
                 notedate_label=NOTEDATE_LABEL, notetext_label=NOTETEXT_LABEL,
                 noteorder_label=None, metadata_labels=None,
                 concepts=None, limit_noteids=None, start_after=0, stop_after=None,
                 select_probability=1.0, **kwargs):
        self.input_files = input_files
        self.package_name = package_name
        self.encoding = encoding
        self.id_label = id_label
        self.noteid_label = noteid_label
        self.notedate_label = notedate_label
        self.notetext_label = notetext_label
        self.noteorder_label = noteorder_label
        self.metadata_labels = metadata_labels
        self.limit_noteids = limit_noteids
        self.start_after = start_after
        self.stop_after = stop_after
        self.select_probability = select_probability
        self.kwargs = kwargs

        self.concepts = list(get_all_concepts(package_name, *(concepts or list())))
        logger.info(f'Loaded {len(self.concepts)} concepts for processing.')

    def run(self, callback):
        """
        callback: function(studyid, note_id, note_date, text, metadata, concept, categories, matches)
        """
        count = 0
        for count, studyid, note_id, note_date, text, metadata in iterate_csv_file(
                self.input_files, encoding=self.encoding,
                id_label=self.id_label, noteid_label=self.noteid_label,
                notedate_label=self.notedate_label, notetext_label=self.notetext_label,
                noteorder_label=self.noteorder_label, metadata_labels=self.metadata_labels,
                start_after=self.start_after, stop_after=self.stop_after,
                select_probability=self.select_probability
        ):
            if self.limit_noteids and note_id not in self.limit_noteids:
                continue
            
            if count % 50000 == 0:
                 logger.info(f'Completed {count:,} records ({datetime.datetime.now()})')

            for concept in self.concepts:
                categories, matches = concept.run_func(text, include_match=True, **metadata)
                callback(studyid, note_id, note_date, text, metadata, concept, categories, matches)
        
        logger.info(f'Finished. Total records: {count:,} ({datetime.datetime.now()})')



"""
Simplify reading input files by creating an iterating wrapper.
"""
import csv
import json
import random
from pathlib import Path

from loguru import logger

from konsepy.constants import NOTEDATE_LABEL, ID_LABEL, NOTEID_LABEL, NOTETEXT_LABEL


class DictReaderInsensitive(csv.DictReader):
    """
    Overrides the csv.fieldnames property making fieldnames stripped and lowercase
    Based on: https://stackoverflow.com/a/16937568/1165522
    """

    @property
    def fieldnames(self):
        fnames = csv.DictReader.fieldnames.fget(self)
        if fnames is None:
            return None
        return [field.strip().lower() for field in fnames]


def iterate_csv_file(input_files, *, start_after=0, stop_after=None,
                     id_label=ID_LABEL, noteid_label=NOTEID_LABEL,
                     notedate_label=NOTEDATE_LABEL, notetext_label=NOTETEXT_LABEL,
                     noteorder_label=None, metadata_labels=None,
                     select_probability=1.0, encoding='latin1'):
    """
    Return count, mrn, note_id, text for each row in csv file

    count: auto-incremented for each record
    """
    count = 0
    total_count = 0
    for input_file in input_files:
        func = None
        if not isinstance(input_file, Path):
            input_file = Path(input_file)
        match input_file.suffix:
            case '.sas7bdat':
                func = _extract_sas_file
            case '.csv':
                func = _extract_csv_file
            case '.tsv':
                func = _extract_csv_file
            case '.jsonl':
                func = _extract_jsonl_file
            case _:
                logger.warning(f'Failed to read corpus file (`input_file`): {input_file}')
                continue
        for mrn, text, note_id, date, md in _deline_lines(
                func, input_file, encoding, id_label, noteid_label,
                notedate_label, notetext_label, noteorder_label,
                metadata_labels):
            if select_probability < 1.0 and random.random() > select_probability:
                continue
            total_count += 1
            if start_after >= total_count:
                continue
            count += 1
            yield count, mrn, note_id, date, text, md
            if stop_after and count > stop_after:
                return


def _deline_lines(func, input_file, encoding, mrn_label, noteid_label,
                  notedate_label, notetext_label, noteorder_label=None,
                  metadata_labels=None):
    # variables for delining notes
    curr_id = None
    curr_mrn = None
    curr_date = None
    curr_doc = []
    for mrn, text, note_id, date, order, md in func(
            input_file, encoding, mrn_label, noteid_label, notedate_label,
            notetext_label, noteorder_label, metadata_labels,
    ):
        if not order:  # skip delining
            yield mrn, text, note_id, date, md
        else:
            if curr_id is None:
                curr_id = note_id
                curr_mrn = mrn
                curr_date = date
            elif note_id != curr_id:
                yield curr_mrn, ' '.join(text for _, _, text in sorted(curr_doc)), curr_id, curr_date, md
                curr_id = note_id
                curr_mrn = mrn
                curr_date = date
                curr_doc = []
            if isinstance(text, float) or not text:
                continue  # skip empty text
            curr_doc.append((note_id, order, text))  # keep track of all text associated with this note
    if curr_id is not None:
        yield curr_mrn, ' '.join(text for _, _, text in sorted(curr_doc)), curr_id, curr_date, md


def _extract_sas_file(input_file, encoding, id_label, noteid_label,
                      notedate_label, notetext_label, noteorder_label=None,
                      metadata_labels=None):
    from sas7bdat import SAS7BDAT
    with SAS7BDAT(input_file, skip_header=False, encoding=encoding) as fh:
        header = []
        for row in fh:
            if not header:
                header = row
                continue
            mrn = row[header.index(id_label)]
            date = row[header.index(notedate_label)] if notedate_label and notedate_label in header else ''
            noteid = row[header.index(noteid_label)]
            text = row[header.index(notetext_label)]
            order = row[header.index(noteorder_label)] if noteorder_label and noteorder_label in header else None
            metadata = {}
            if metadata_labels:
                for src, (dest, func) in metadata_labels.items():
                    metadata[dest] = func(row[header.index(src)])
            yield mrn, text, noteid, date, order, metadata


def _extract_csv_file(input_file, encoding, id_label, noteid_label, notedate_label,
                      notetext_label, noteorder_label=None, metadata_labels=None):
    with open(input_file, newline='', encoding=encoding) as fh:
        for row in DictReaderInsensitive(fh):
            text = row[notetext_label]
            mrn = row[id_label]
            date = row.get(notedate_label, '')
            note_id = row[noteid_label]
            order = row.get(noteorder_label, '')
            metadata = {}
            if metadata_labels:
                for src, (dest, func) in metadata_labels.items():
                    metadata[dest] = func(row[src])
            yield mrn, text, note_id, date, order, metadata


def _extract_jsonl_file(input_file, encoding, id_label, noteid_label, notedate_label,
                        notetext_label, noteorder_label=None, metadata_labels=None):
    with open(input_file, encoding=encoding) as fh:
        for line in fh:
            data = json.loads(line.strip())
            text = data[notetext_label]
            mrn = data[id_label]
            date = data.get(notedate_label, '')
            note_id = data[noteid_label]
            order = data.get(noteorder_label, '')
            metadata = {}
            if metadata_labels:
                for src, (dest, func) in metadata_labels.items():
                    metadata[dest] = func(data[src])
            yield mrn, text, note_id, date, order, metadata


def output_results(outdir, *, not_found_text=None,
                   note_counter=None, cat_counter_mrns=None,
                   category_enums=None, note_to_cat=None, mrn_to_cat=None):
    categories = [str(e) for category_enum in category_enums for e in category_enum]
    if not_found_text is not None:
        with open(outdir / 'snippets.csv', 'w', newline='') as out:
            writer = csv.writer(out)
            writer.writerow(['count', 'snippet'])
            for snippet, count in not_found_text.most_common():
                writer.writerow([count, ' '.join(snippet.split())])

    with open(outdir / 'category_counts.csv', 'w', newline='') as out:
        writer = csv.writer(out)
        writer.writerow(['category', 'note_count', 'mrn_count'])
        for cat in categories:
            writer.writerow([cat, note_counter[cat], len(cat_counter_mrns[cat])])

    with open(outdir / 'mrn_category_counts.csv', 'w', newline='') as out:
        writer = csv.DictWriter(out, ['mrn'] + categories)
        writer.writeheader()
        for mrn, note_counter in mrn_to_cat.items():
            writer.writerow({'mrn': mrn} | dict(note_counter))

    with open(outdir / 'notes_category_counts.csv', 'w', newline='') as out:
        writer = csv.DictWriter(out, ['mrn', 'note_id'] + categories)
        writer.writeheader()
        for (mrn, note), note_counter in note_to_cat.items():
            writer.writerow({'mrn': mrn, 'note_id': note} | dict(note_counter))
    logger.info(f'Unique MRNs with any category: {len(mrn_to_cat):,}')
